#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
NTU Condenser model, for use with Helmholtz EOS property packages.
"""

__author__ = "Jinliang Ma"

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue
# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util import get_solver, from_json, to_json, StoreSpec
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError
from idaes.generic_models.unit_models.heater import (
    _make_heater_config_block, _make_heater_control_volume)
import idaes.core.util.unit_costing as costing
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


def _make_heat_exchanger_config(config):
    """
    Declare configuration options for HeatExchangerData block.
    """
    config.declare(
        "hot_side_name",
        ConfigValue(
            default="shell",
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "cold_side_name",
        ConfigValue(
            default="tube",
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "hot_side_config",
        ConfigBlock(
            implicit=True,
            description="Config block for hot side",
            doc="""A config block used to construct the hot side control volume.
This config can be given by the hot side name instead of hot_side_config.""",
        ),
    )
    config.declare(
        "cold_side_config",
        ConfigBlock(
            implicit=True,
            description="Config block for cold side",
            doc="""A config block used to construct the cold side control volume.
This config can be given by the cold side name instead of cold_side_config.""",
        ),
    )
    _make_heater_config_block(config.hot_side_config)
    _make_heater_config_block(config.cold_side_config)


@declare_process_block_class("HelmNtuCondenser", doc="Simple 0D condenser model.")
class HelmNtuCondenserData(UnitModelBlockData):
    """
    Simple NTU condenser unit model.  This model assumes the property pacakages
    specified are Helmholtz EOS type.
    """
    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    _make_heat_exchanger_config(CONFIG)

    def _process_config(self):
        """Check for configuration errors and alternate config option names.
        """
        config = self.config

        if config.hot_side_name == config.cold_side_name:
            raise NameError(
                "Condenser hot and cold side cannot have the same name '{}'."
                " Be sure to set both the hot_side_name and cold_side_name.".format(
                    config.hot_side_name
                )
            )
        for o in config:
            if not (
                o in self.CONFIG or o in [config.hot_side_name, config.cold_side_name]
            ):
                raise KeyError("Condenser config option {} not defined".format(o))

        if config.hot_side_name in config:
            config.hot_side_config.set_value(config[config.hot_side_name])
            # Allow access to hot_side_config under the hot_side_name
            setattr(config, config.hot_side_name, config.hot_side_config)
        if config.cold_side_name in config:
            config.cold_side_config.set_value(config[config.cold_side_name])
            # Allow access to hot_side_config under the cold_side_name
            setattr(config, config.cold_side_name, config.cold_side_config)

        if config.cold_side_name in ["hot_side", "side_1"]:
            raise ConfigurationError("Cold side name cannot be in ['hot_side', 'side_1'].")
        if config.hot_side_name in ["cold_side", "side_2"]:
            raise ConfigurationError("Hot side name cannot be in ['cold_side', 'side_2'].")

    def build(self):
        """
        Building model

        Args:
            None
        Returns:
            None
        """
        ########################################################################
        #  Call UnitModel.build to setup dynamics and configure                #
        ########################################################################
        super().build()
        self._process_config()
        config = self.config
        time = self.flowsheet().time

        ########################################################################
        # Add control volumes                                                  #
        ########################################################################
        hot_side = _make_heater_control_volume(
            self,
            config.hot_side_name,
            config.hot_side_config,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        cold_side = _make_heater_control_volume(
            self,
            config.cold_side_name,
            config.cold_side_config,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        # Add refernces to the hot side and cold side, so that we have solid
        # names to refere to internally.  side_1 and side_2 also maintain
        # compatability with older models.  Using add_object_reference keeps
        # these from showing up when you iterate through pyomo compoents in a
        # model, so only the user specified control volume names are "seen"
        if not hasattr(self, "side_1"):
            add_object_reference(self, "side_1", hot_side)
        if not hasattr(self, "side_2"):
            add_object_reference(self, "side_2", cold_side)
        if not hasattr(self, "hot_side"):
            add_object_reference(self, "hot_side", hot_side)
        if not hasattr(self, "cold_side"):
            add_object_reference(self, "cold_side", cold_side)

        ########################################################################
        # Add variables                                                        #
        ########################################################################
        # Use hot side units as basis
        s1_metadata = config.hot_side_config.property_package.get_metadata()

        f_units = s1_metadata.get_derived_units("flow_mole")
        cp_units = s1_metadata.get_derived_units("heat_capacity_mole")
        q_units = s1_metadata.get_derived_units("power")
        u_units = s1_metadata.get_derived_units("heat_transfer_coefficient")
        a_units = s1_metadata.get_derived_units("area")
        temp_units = s1_metadata.get_derived_units("temperature")

        self.overall_heat_transfer_coefficient = pyo.Var(
            time,
            domain=pyo.PositiveReals,
            initialize=100.0,
            doc="Overall heat transfer coefficient",
            units=u_units,
        )
        self.area = pyo.Var(
            domain=pyo.PositiveReals,
            initialize=1000.0,
            doc="Heat exchange area",
            units=a_units,
        )
        self.heat_duty = pyo.Reference(cold_side.heat)
        ########################################################################
        # Add ports                                                            #
        ########################################################################
        i1 = self.add_inlet_port(
            name=f"{config.hot_side_name}_inlet",
            block=hot_side,
            doc="Hot side inlet")
        i2 = self.add_inlet_port(
            name=f"{config.cold_side_name}_inlet",
            block=cold_side,
            doc="Cold side inlet")
        o1 = self.add_outlet_port(
            name=f"{config.hot_side_name}_outlet",
            block=hot_side,
            doc="Hot side outlet")
        o2 = self.add_outlet_port(
            name=f"{config.cold_side_name}_outlet",
            block=cold_side,
            doc="Cold side outlet")

        # Using Andrew's function for now.  I want these port names for backward
        # compatablity, but I don't want them to appear if you iterate throught
        # components and add_object_reference hides them from Pyomo.
        if not hasattr(self, "inlet_1"):
            add_object_reference(self, "inlet_1", i1)
        if not hasattr(self, "inlet_2"):
            add_object_reference(self, "inlet_2", i2)
        if not hasattr(self, "outlet_1"):
            add_object_reference(self, "outlet_1", o1)
        if not hasattr(self, "outlet_2"):
            add_object_reference(self, "outlet_2", o2)

        if not hasattr(self, "hot_inlet"):
            add_object_reference(self, "hot_inlet", i1)
        if not hasattr(self, "cold_inlet"):
            add_object_reference(self, "cold_inlet", i2)
        if not hasattr(self, "hot_outlet"):
            add_object_reference(self, "hot_outlet", o1)
        if not hasattr(self, "cold_outlet"):
            add_object_reference(self, "cold_outlet", o2)

        ########################################################################
        # Add a unit level energy balance                                      #
        ########################################################################
        @self.Constraint(time, doc="Heat balance equation")
        def unit_heat_balance(b, t):
            return 0 == (hot_side.heat[t] +
                pyunits.convert(cold_side.heat[t], to_units=q_units))

        ########################################################################
        # Add some useful expressions for condenser performance                #
        ########################################################################

        @self.Expression(time, doc="Inlet temperature difference")
        def delta_temperature_in(b, t):
                return (hot_side.properties_in[t].temperature -
                    pyunits.convert(cold_side.properties_in[t].temperature, temp_units))

        @self.Expression(time, doc="Outlet temperature difference")
        def delta_temperature_out(b, t):
                return (hot_side.properties_out[t].temperature -
                    pyunits.convert(cold_side.properties_out[t].temperature, temp_units))

        @self.Expression(time, doc="NTU Based temperature difference")
        def delta_temperature_ntu(b, t):
                return (hot_side.properties_in[t].temperature_sat -
                    pyunits.convert(cold_side.properties_in[t].temperature, temp_units))

        @self.Expression(time, doc="Minimum product of flow rate and heat "
            "capacity (always on tube side since shell side has phase change)")
        def mcp_min(b, t):
            return pyunits.convert(cold_side.properties_in[t].flow_mol
                * cold_side.properties_in[t].cp_mol_phase['Liq'], f_units*cp_units)

        @self.Expression(time, doc="Number of transfer units (NTU)")
        def ntu(b, t):
            return b.overall_heat_transfer_coefficient[t]*b.area/b.mcp_min[t]

        @self.Expression(time, doc="Condenser effectiveness factor")
        def effectiveness(b, t):
            return 1 - pyo.exp(-self.ntu[t])

        @self.Expression(time, doc="Heat treansfer")
        def heat_transfer(b, t):
            return b.effectiveness[t]*b.mcp_min[t] * b.delta_temperature_ntu[t]

        ########################################################################
        # Add Equations to calculate heat duty based on NTU method             #
        ########################################################################
        @self.Constraint(
            time, doc="Heat transfer rate equation based on NTU method")
        def heat_transfer_equation(b, t):
            return (pyunits.convert(cold_side.heat[t], q_units) ==
                self.heat_transfer[t])

        @self.Constraint(
            time, doc="Shell side outlet enthalpy is saturated water enthalpy")
        def saturation_eqn(b, t):
            return (hot_side.properties_out[t].enth_mol ==
                hot_side.properties_in[t].enth_mol_sat_phase["Liq"])

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.hot_side.material_accumulation[:, :, :].value = 0
            self.hot_side.energy_accumulation[:, :].value = 0
            self.hot_side.material_accumulation[0, :, :].fix(0)
            self.hot_side.energy_accumulation[0, :].fix(0)
            self.cold_side.material_accumulation[:, :, :].value = 0
            self.cold_side.energy_accumulation[:, :].value = 0
            self.cold_side.material_accumulation[0, :, :].fix(0)
            self.cold_side.energy_accumulation[0, :].fix(0)

    def initialize(
        self,
        state_args_1=None,
        state_args_2=None,
        unfix='hot_flow',
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Condenser initialization method. The initialization routine assumes
        fixed area and heat transfer coefficient and adjusts the cooling water
        flow to condense steam to saturated water at shell side pressure.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for hot side (see documentation of the specific
                property package) (default = None).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for cold side (see documentation of the specific
                property package) (default = None).
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None

        """
        if unfix not in {"hot_flow", "cold_flow", "pressure"}:
            raise Exception("Condenser free variable must be in 'hot_flow', "
                "'cold_flow', or 'pressure'")
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        hot_side = getattr(self, self.config.hot_side_name)
        cold_side = getattr(self, self.config.cold_side_name)

        # Store initial model specs, restored at the end of initializtion, so
        # the problem is not altered.  This can restore fixed/free vars,
        # active/inactive constraints, and fixed variable values.
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1)
        flags2 = cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2)
        init_log.info_high("Initialization Step 1 Complete.")

        # Solve with all constraints activated
        self.saturation_eqn.activate()
        if unfix == 'pressure':
            hot_side.properties_in[:].pressure.unfix()
        elif unfix == 'hot_flow':
            hot_side.properties_in[:].flow_mol.unfix()
        elif unfix == 'cold_flow':
            cold_side.properties_in[:].flow_mol.unfix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high(
            "Initialization Step 4 {}.".format(idaeslog.condition(res))
        )

        # Release Inlet state
        hot_side.release_state(flags1, outlvl)
        cold_side.release_state(flags2, outlvl)
        from_json(self, sd=istate, wts=sp)


    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "HX Coefficient": self.overall_heat_transfer_coefficient[time_point]
        }
        var_dict["HX Area"] = self.area
        var_dict["Heat Duty"] = self.heat_duty[time_point]
        expr_dict = {}
        expr_dict["Delta T In"] = self.delta_temperature_in[time_point]
        expr_dict["Delta T Out"] = self.delta_temperature_out[time_point]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot Inlet": self.hot_inlet,
                "Hot Outlet": self.hot_outlet,
                "Cold Inlet": self.cold_inlet,
                "Cold Outlet": self.cold_outlet,
            },
            time_point=time_point,
        )

    def get_costing(self, module=costing):
        if not hasattr(self.flowsheet(), "costing"):
            self.flowsheet().get_costing()
        self.costing = pyo.Block()

        module.hx_costing(self.costing)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        area_sf_default = 1e-2
        overall_heat_transfer_coefficient_sf_default = 1e-2

        # Function to set defaults so I don't need to reproduce the same code
        def _fill_miss_with_default(name, s):
            try:
                c = getattr(self, name)
            except AttributeError:
                return # it's okay if the attribute doesn't exist, spell careful
            if iscale.get_scaling_factor(c) is None:
                for ci in c.values():
                    if iscale.get_scaling_factor(ci) is None:
                        iscale.set_scaling_factor(ci, s)

        # Set defaults where scale factors are missing
        _fill_miss_with_default("area", area_sf_default)
        _fill_miss_with_default(
            "overall_heat_transfer_coefficient",
             overall_heat_transfer_coefficient_sf_default
        )

        for t, c in self.heat_transfer_equation.items():
            sf = iscale.get_scaling_factor(self.cold_side.heat[t])
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.unit_heat_balance.items():
            sf = iscale.get_scaling_factor(self.cold_side.heat[t])
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.saturation_eqn.items():
            sf = iscale.get_scaling_factor(
                self.hot_side.properties_out[t].enth_mol)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import os
import hashlib
import idaes.logger as idaeslog
import tarfile
import idaes
from shutil import copyfile
from pyomo.common.download import FileDownloader
import urllib

_log = idaeslog.getLogger(__name__)
_release_base_url = idaes.config.release_base_url


class UnsupportedPlatformError(RuntimeError):
    pass


def hash_file_sha256(fname):
    """Calculate sha256 hash of potentially large files.

    Args:
        fname: path to file to hash

    Returns:
        hash as a string
    """
    with open(fname, 'rb') as f:
        fh = hashlib.sha256()
        while True:
            fb = f.read(10000)
            if len(fb) <= 0:
                break
            fh.update(fb)
    return str(fh.hexdigest())

_hash = hash_file_sha256

def _get_file_downloader(insecure, cacert):
    fd = FileDownloader(insecure=insecure, cacert=cacert)
    arch = fd.get_sysinfo()
    if arch[1] != 64:
        _log.error("IDAES Extensions only supports 64bit Python.")
        raise RuntimeError("IDAES Extensions only supports 64bit Python.")
    return fd, arch


def _get_platform(fd, platform, arch):
    if platform == "auto":
        platform = arch[0]
    if platform == "linux":
        linux_dist = fd.get_os_version().replace(".", "")
        _log.debug(f"Detected Linux distribution: {linux_dist}")
        if linux_dist in idaes.config.known_binary_platform:
            platform = linux_dist
        else:
            raise UnsupportedPlatformError(
                f"Detected platform {linux_dist} is not recognized as "
                "supported platform.")
    if platform not in idaes.config.known_binary_platform:
        raise UnsupportedPlatformError(f"Unknown platform: {platform}.")
    if platform in idaes.config.binary_platform_map:
        platform = idaes.config.binary_platform_map[platform]
    _log.debug(f"Downloading binaries for {platform}")
    return platform


def _get_checksums(fd, to_path, release):
    checksum = {} # storage for hashes if install from release
    check_to = os.path.join(to_path, f"sha256sum_{release}.txt")
    check_from = idaes.config.release_checksum_url.format(release)
    _log.debug(f"Getting checksum file {check_from}")
    fd.set_destination_filename(check_to)
    fd.get_binary_file(check_from)
    # read the hashes file and store then in checksum dict
    with open(check_to, 'r') as f:
        for i in range(1000):
            line = f.readline(1000)
            if line == "":
                break
            line = line.split(sep="  ")
            checksum[line[1].strip()] = line[0].strip()
    return checksum


def _get_release_url_and_checksum(fd, to_path, release, url, nochecksum):
    checksum = False # default if not checking checksums
    if release is not None:
        url = "/".join([_release_base_url, release])
        if not nochecksum:
            checksum = _get_checksums(fd, to_path, release)
        else:
            _log.debug("Skip release checksum verification at user request.")
    else:
        _log.debug("Release not specified.")
    if url is None:
        _log.debug("No release or URL was provided.")
        raise Exception("Must provide a location to download binaries")
    if url.endswith("/"):
        url = url[0:-1] # if url ends with "/" remove it for proper join later
    _log.debug(f"Downloading binaries from {url}")
    return url, checksum


def _download_package(fd, name, frm, to, platform):
    _log.debug(f"Getting {name} from: {frm}")
    _log.debug(f"Saving solvers to: {to}")
    fd.set_destination_filename(to)
    try:
        fd.get_binary_file(frm)
    except urllib.error.HTTPError:
        raise Exception(f"{name} binaries are unavailable for {platform}")


def download_binaries(
    release=None,
    url=None,
    insecure=False,
    cacert=None,
    verbose=False,
    platform="auto",
    nochecksum=False,
    library_only=False,
    no_download=False,
    extras_only=False,
    extra=(),
    to_path=None,
    alt_path=None,
    ):
    """
    Download IDAES solvers and libraries and put them in the right location. Need
    to supply either local or url argument.

    Args:
        url (str): a url to download binary files to install files
        to_path: for testing only, an alternate subdirectory of the idaes data
                 directory for a test binary install.
    Returns:
        None
    """
    if verbose:
        _log.setLevel(idaeslog.DEBUG)
    if no_download:
        nochecksum = True
    # set the locations to download files to, to_path is an alternate
    # subdirectory of idaes.data_directory that can optionally be used to test
    # this function without interfereing with anything else.  It's a subdirectory
    # of the data directory because that should be a safe place to store some
    # test files.
    if alt_path is not None:
        to_path = os.path.abspath(alt_path)
    if to_path is None:
        to_path = idaes.bin_directory
    else:
        to_path = os.path.join(idaes.data_directory, to_path)
    idaes._create_bin_dir(to_path)
    fd, arch = _get_file_downloader(insecure, cacert)
    platform = _get_platform(fd, platform, arch)
    url, checksum = _get_release_url_and_checksum(
        fd, to_path, release, url, nochecksum)
    # Set the binary file destinations
    pname = []
    ptar = []
    ftar = []
    furl = []
    def _add_pack(name):
        f = f"idaes-{name}-{platform}-{arch[1]}.tar.gz"
        ftar.append(f)
        ptar.append(os.path.join(to_path, f))
        pname.append(name)
        furl.append("/".join([url, f]))

    for e in extra:
        _add_pack(e) # you have to explicitly ask for extras so assume you want
    if not extras_only:
        _add_pack("lib")
    if not library_only and not extras_only:
        _add_pack("solvers")

    if no_download:
        d = {
            "release": release,
            "platform": platform,
            "bits": arch[1],
        }
        for n, p, u in zip(pname, ptar, furl):
            d[u] = p
        return d

    # download packages
    for n, p, u in zip(pname, ptar, furl):
        _download_package(fd, n, frm=u, to=p, platform=platform)

    # If release checksum is not empty, nochecksum opt allows hash to be ignored
    if checksum:
        for n, p, f in zip(pname, ptar, ftar):
            hash_l = _hash(p)
            _log.debug(f"{n} Hash {hash_l}")
            if checksum.get(f, "") != hash_l:
                raise Exception(f"{n} hash does not match expected")

    # Extract solvers
    for n, p in zip(pname, ptar):
        _log.debug(f"Extracting files in {p} to {to_path}")
        with tarfile.open(p, 'r') as f:
            f.extractall(to_path)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for the IDAES Flowsheet Visualizer (IFV).

These are currently integration tests, because the start/stop the embedded HTTP server.
"""
import glob
import json
import logging
import os
from pathlib import Path
import pytest
import re
import requests
import time

from pyomo.environ import ConcreteModel, SolverFactory, Constraint, value
from idaes.core import FlowsheetBlock
from idaes.generic_models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.generic_models.unit_models import Flash
from idaes.ui.fsvis import fsvis, errors
from idaes.ui.flowsheet import validate_flowsheet


@pytest.fixture(scope="module")
def flash_model():
    """Flash unit model. Use '.fs' attribute to get the flowsheet."""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    # Flash properties
    m.fs.properties = BTXParameterBlock(
        default={
            "valid_phase": ("Liq", "Vap"),
            "activity_coeff_model": "Ideal",
            "state_vars": "FTPz",
        }
    )
    # Flash unit
    m.fs.flash = Flash(default={"property_package": m.fs.properties})
    m.fs.flash.inlet.flow_mol.fix(1)
    m.fs.flash.inlet.temperature.fix(368)
    m.fs.flash.inlet.pressure.fix(101325)
    m.fs.flash.inlet.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.flash.inlet.mole_frac_comp[0, "toluene"].fix(0.5)
    m.fs.flash.heat_duty.fix(0)
    m.fs.flash.deltaP.fix(0)
    return m


@pytest.mark.integration
def test_visualize(flash_model, tmp_path):
    from pathlib import Path

    flowsheet = flash_model.fs
    # Start the visualization server
    result = fsvis.visualize(flowsheet, "Flash", browser=False, save_dir=tmp_path)
    # Get the model
    resp = requests.get(f"http://127.0.0.1:{result.port}/fs?id=Flash")
    data = resp.json()
    # Validate the model
    ok, msg = validate_flowsheet(data)
    assert ok, f"Invalid flowsheet returned: {msg}"
    assert data["model"]["id"] == "Flash"
    assert data["model"]["unit_models"]["flash"]["type"] == "flash"
    assert len(data["cells"]) == 7
    units = [x for x in data["cells"] if x["type"] == "standard.Image"]
    assert len(units) == 4
    unit_images = [Path(x["attrs"]["image"]["xlinkHref"]).name for x in units]
    unit_images.sort()
    assert unit_images == ["feed.svg", "flash.svg", "product.svg", "product.svg"]
    # Modify the model by deleting its one and only component
    flowsheet.del_component("flash")
    # Get the model (again)
    resp = requests.get(f"http://127.0.0.1:{result.port}/fs?id=Flash")
    data = resp.json()
    # Validate the modified model
    expected = {
        "model": {
            "id": "Flash",
            "stream_table": {"columns": ["", "Variable"], "data": [], "index": []},
            "unit_models": {},
            "arcs": {},
        },
        "cells": [],
    }
    assert data == expected


@pytest.mark.integration
def test_save_visualization(flash_model, tmp_path):
    # view logs from the persistence module
    logging.getLogger("idaes.ui.fsvis").setLevel(logging.DEBUG)
    flowsheet = flash_model.fs
    # Start the visualization server, using temporary save location
    save_location = tmp_path / "flash-vis.json"
    fsvis_result = fsvis.visualize(
        flowsheet, "Flash", browser=False, save=save_location, save_dir=tmp_path
    )
    # Check the contents of the saved file are the same as what is returned by the server
    with open(fsvis_result.store.filename) as fp:
        file_data = json.load(fp)
    resp = requests.get(f"http://127.0.0.1:{fsvis_result.port}/fs?id=Flash")
    net_data = resp.json()
    assert file_data == net_data


def _canonicalize(d):
    for cell in d["cells"]:
        if "ports" in cell:
            items = cell["ports"]["items"]
            cell["ports"]["items"] = sorted(items, key=lambda x: x["group"])


@pytest.mark.unit
def test_invoke(flash_model):
    # from inspect import signature -- TODO: use for checking params
    from idaes.ui import fsvis as fsvis_pkg

    functions = {
        "method": getattr(flash_model.fs, "visualize"),
        "package": getattr(fsvis_pkg, "visualize"),
        "module": getattr(fsvis, "visualize"),
    }
    # TODO: check params


@pytest.mark.unit
def test_visualize_fn(flash_model):
    flowsheet = flash_model.fs
    result = fsvis.visualize(flowsheet, browser=False, save=False)
    assert result.store.filename == ""
    #
    for bad_save_as in (1, "/no/such/file/exists.I.hope", flowsheet):
        with pytest.raises(errors.VisualizerError):
            fsvis.visualize(flowsheet, save=bad_save_as, browser=False)


@pytest.mark.unit
def test_flowsheet_name(flash_model, tmp_path):
    raw_name = "Hello World"
    result = fsvis.visualize(
        flash_model.fs, name=raw_name, browser=False, save_dir=tmp_path
    )
    assert re.search(raw_name, result.store.filename)


@pytest.mark.unit
def test_mock_webbrowser(flash_model):
    from idaes.ui.fsvis import fsvis

    wb = fsvis.webbrowser
    for wb_mock in (MockWB(True), MockWB(False)):
        fsvis.webbrowser = wb_mock
        _ = fsvis.visualize(flash_model.fs, save=False)
    fsvis.webbrowser = wb


class MockWB:
    def __init__(self, ok):
        self.ok = ok

    def open(self, *args):
        return self.ok


## Test saving of the status file


@pytest.fixture
def save_files_prefix(tmp_path):
    value = str(tmp_path / "test_visualize")
    # clear out any cruft
    for filename in glob.glob(str(tmp_path / "test_visualize*")):
        os.unlink(filename)
    yield value
    # clear out any cruft (2)
    for filename in glob.glob(str(tmp_path / "test_visualize*")):
        os.unlink(filename)


@pytest.mark.unit
def test_visualize_save_versions(flash_model, save_files_prefix):
    # test versioned file saves
    flowsheet = flash_model.fs
    path = Path(save_files_prefix + "_save")
    work_dir = path.parent
    fs_name = path.name
    for i in range(4):
        save_arg = (True, None)[i % 2]  # try both kinds of 'use default' values
        if i < 3:
            result = fsvis.visualize(
                flowsheet, fs_name, save_dir=work_dir, browser=False, save=save_arg
            )
            if i == 0:
                assert re.search(f"{path.name}.json", result.store.filename)
            else:
                assert re.search(f"{path.name}.*{i}.*\.json", result.store.filename)
        else:
            msv, fsvis.MAX_SAVED_VERSIONS = fsvis.MAX_SAVED_VERSIONS, i - 1
            with pytest.raises(RuntimeError):
                fsvis.visualize(flowsheet, fs_name, save_dir=work_dir, browser=False)
            fsvis.MAX_SAVED_VERSIONS = msv


@pytest.mark.unit
def test_visualize_save_explicit(flash_model, save_files_prefix):
    # test explicit filename
    flowsheet = flash_model.fs
    howdy = Path(save_files_prefix + "_howdy")
    result = fsvis.visualize(flowsheet, "flowsheet", save=howdy, browser=False)
    assert re.search(howdy.name, result.store.filename)
    # overwrite but this time break explicit file into relative name and directory
    result = fsvis.visualize(
        flowsheet,
        "flowsheet",
        save=howdy.name,
        save_dir=howdy.parent,
        browser=False,
        overwrite=True,
    )
    assert re.search(howdy.name, result.store.filename)


@pytest.mark.unit
def test_visualize_save_cannot(flash_model, tmp_path):
    flowsheet = flash_model.fs
    with pytest.raises(errors.VisualizerError):
        fsvis.visualize(flowsheet, "foo", save="foo", save_dir=Path("/a/b/c/d/e/f/g"))


@pytest.mark.unit
def test_visualize_save_overwrite(flash_model, save_files_prefix):
    flowsheet = flash_model.fs
    howdy = Path(save_files_prefix + "_howdy")
    howdy.open("w").write("howdy")
    howdy_stat = os.stat(howdy)
    result = fsvis.visualize(
        flowsheet, "flowsheet", save=howdy, overwrite=True, browser=False
    )
    howdy_stat2 = os.stat(result.store.filename)
    assert (
        howdy_stat2.st_mtime > howdy_stat.st_mtime
    )  # modification time should be later


@pytest.mark.unit
def test_pick_default_save_location():
    from idaes.ui.fsvis.fsvis import _pick_default_save_location as pdsl
    p = pdsl("foo", None)
    assert str(p).endswith("foo.json")
    p = pdsl("foo", Path("/a"))
    assert p == Path("/a") / "foo.json"


@pytest.mark.unit
def test_existing_save_path(tmp_path):
    from idaes.ui.fsvis.fsvis import _handle_existing_save_path as hesp
    name = "foo"
    save_path = tmp_path / (name + ".json")
    # not there
    p = hesp(name, save_path)
    assert p == save_path
    # version 1
    save_path.open("w").write("hello")
    p1 = hesp(name, save_path)
    assert p1 != save_path
    # version 2
    p1.open("w").write("hello")
    p2 = hesp(name, save_path)
    assert str(p2) > str(p1)
    # version too far
    p2.open("w").write("hello")
    with pytest.raises(errors.TooManySavedVersions):
        p3 = hesp(name, save_path, max_versions=2)
    # infinite versions
    p4 = hesp(name, save_path, max_versions=0)
    assert str(p4) > str(p2)
    # overwrite
    p0 = hesp(name, save_path, overwrite=True)
    assert p0 == save_path


@pytest.mark.component
def test_loop_forever():
    from threading import Thread
    for quietness in (True, False):
        thr = Thread(target=fsvis._loop_forever, args=(quietness,))
        thr.setDaemon(True)
        thr.start()
        # wait a while, make sure it's still alive
        time.sleep(3)
        assert thr.is_alive()
    # threads should die when process exits



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import pyomo.common.unittest as unittest
import pytest
from idaes.surrogate.pysmo.utils import NumpyEvaluator
from pyomo.environ import ConcreteModel, Param, Var, value, sin, atan, atanh
from pyomo.core import ComponentMap

try:
    import numpy as np
    _numpy_available = True
except ImportError:
    _numpy_available = False

@unittest.skipIf(not _numpy_available, "Test requires numpy")
class TestNumpyEvaluator:

    @pytest.mark.unit
    def test_eval_numpy(self):
        m = ConcreteModel()
        m.p = Param([1,2], mutable=True)   
        m.x = Var()

        data = np.array([[0,-1,2],
                         [.1,.2,.3],
                         [4,5,6]])
        cMap = ComponentMap()
        cMap[m.p[1]] = data[0]
        cMap[m.p[2]] = data[1]
        cMap[m.x] = data[2]

        npe = NumpyEvaluator(cMap)

        result = npe.walk_expression(sin(m.x))
        assert result[0] == sin(4)
        assert result[1] == sin(5)
        assert result[2] == sin(6)

        result = npe.walk_expression(abs(m.x * m.p[1] - m.p[2]))
        assert result[0] == .1
        assert result[1] == -((-1*5)-.2)
        assert result[2] == (2*6-.3)

        result = npe.walk_expression(atan(m.x))
        assert result[0] == atan(4)
        assert result[1] == atan(5)
        assert result[2] == atan(6)

        result = npe.walk_expression(atanh(m.p[2]))
        assert result[0] == atanh(.1)
        assert result[1] == atanh(.2)
        assert result[2] == atanh(.3)

    @pytest.mark.unit
    def test_eval_constant(self):
        m = ConcreteModel()
        m.p = Param([1,2], mutable=True)
        m.x = Var(initialize=0.25)

        cMap = ComponentMap()
        cMap[m.p[1]] = 2
        cMap[m.p[2]] = 4

        npe = NumpyEvaluator(cMap)

        expr = m.p[1] + m.p[2] + m.x + .5
        assert npe.walk_expression(expr) == 6.75

        m.p[1] = 2
        m.p[2] = 4
        assert value(expr) == 6.75

if __name__ == "__main__":
    pytest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for calculating equilibrium constants
"""
from pyomo.environ import exp, Var, units as pyunits, value

from idaes.generic_models.properties.core.generic.generic_reaction import \
    ConcentrationForm
from .dh_rxn import constant_dh_rxn
from idaes.core.util.misc import set_param_from_config
from idaes.core.util.constants import Constants as c
from idaes.core.util.exceptions import BurntToast, ConfigurationError


# -----------------------------------------------------------------------------
# Constant Keq
class ConstantKeq():

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name))
        elif (c_form == ConcentrationForm.moleFraction or
              c_form == ConcentrationForm.massFraction or
              c_form == ConcentrationForm.activity):
            e_units = pyunits.dimensionless
        else:
            order = 0

            try:
                # This will work for Reaction Packages
                pc_set = parent.config.property_package._phase_component_set
            except AttributeError:
                # Need to allow for inherent reactions in Property Packages
                if not parent._electrolyte:
                    # In most cases ,should have _phase_component_set
                    pc_set = parent._phase_component_set
                else:
                    # However, for electrolytes need true species set
                    pc_set = parent.true_phase_component_set

            for p, j in pc_set:
                order += rblock.reaction_order[p, j].value

            if c_form == ConcentrationForm.molarity:
                c_units = units["density_mole"]
            elif c_form == ConcentrationForm.molality:
                c_units = units["amount"]*units["mass"]**-1
            elif c_form == ConcentrationForm.partialPressure:
                c_units = units["pressure"]
            else:
                raise BurntToast(
                    "{} get_concentration_term received unrecognised "
                    "ConcentrationForm ({}). This should not happen - please "
                    "contact the IDAES developers with this bug."
                    .format(rblock.name, c_form))

            e_units = c_units**order

        rblock.k_eq_ref = Var(
                doc="Equilibrium constant at reference state",
                units=e_units)
        set_param_from_config(rblock, param="k_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        return rblock.k_eq_ref

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        return 1/value(rblock.k_eq_ref)


# -----------------------------------------------------------------------------
# van t'Hoff equation (constant dh_rxn)
class van_t_hoff():

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name))
        elif (c_form == ConcentrationForm.moleFraction or
              c_form == ConcentrationForm.massFraction or
              c_form == ConcentrationForm.activity):
            e_units = pyunits.dimensionless
        else:
            order = 0

            try:
                # This will work for Reaction Packages
                pc_set = parent.config.property_package._phase_component_set
            except AttributeError:
                # Need to allow for inherent reactions in Property Packages
                if not parent._electrolyte:
                    # In most cases ,should have _phase_component_set
                    pc_set = parent._phase_component_set
                else:
                    # However, for electrolytes need true species set
                    pc_set = parent.true_phase_component_set

            for p, j in pc_set:
                order += rblock.reaction_order[p, j].value

            if c_form == ConcentrationForm.molarity:
                c_units = units["density_mole"]
            elif c_form == ConcentrationForm.molality:
                c_units = units["amount"]*units["mass"]**-1
            elif c_form == ConcentrationForm.partialPressure:
                c_units = units["pressure"]
            else:
                raise BurntToast(
                    "{} get_concentration_term received unrecognised "
                    "ConcentrationForm ({}). This should not happen - please "
                    "contact the IDAES developers with this bug."
                    .format(rblock.name, c_form))

            e_units = c_units**order

        rblock.k_eq_ref = Var(
                doc="Equilibrium constant at reference state",
                units=e_units)
        set_param_from_config(rblock, param="k_eq_ref", config=config)

        rblock.T_eq_ref = Var(
                doc="Reference temperature for equilibrium constant",
                units=units["temperature"])
        set_param_from_config(rblock, param="T_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        units = rblock.parent_block().get_metadata().derived_units

        return rblock.k_eq_ref * exp(
            -(b.dh_rxn[r_idx] /
              pyunits.convert(c.gas_constant,
                              to_units=units["gas_constant"])) *
            (1/T - 1/rblock.T_eq_ref))

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        return 1/value(rblock.k_eq_ref)


# -----------------------------------------------------------------------------
# Constant dh_rxn and ds_rxn
class gibbs_energy():

    @staticmethod
    def build_parameters(rblock, config):
        parent = rblock.parent_block()
        units = parent.get_metadata().derived_units

        c_form = config.concentration_form
        if c_form is None:
            raise ConfigurationError(
                "{} concentration_form configuration argument was not set. "
                "Please ensure that this argument is included in your "
                "configuration dict.".format(rblock.name))
        elif (c_form == ConcentrationForm.molarity or
              c_form == ConcentrationForm.activity):

            order = 0
            try:
                # This will work for Reaction Packages
                pc_set = parent.config.property_package._phase_component_set
            except AttributeError:
                # Need to allow for inherent reactions in Property Packages
                if not parent._electrolyte:
                    # In most cases ,should have _phase_component_set
                    pc_set = parent._phase_component_set
                else:
                    # However, for electrolytes need true species set
                    pc_set = parent.true_phase_component_set

            for p, j in pc_set:
                order += rblock.reaction_order[p, j].value

            rblock._keq_units = (pyunits.convert(1*pyunits.mol/pyunits.L,
                                                 units["density_mole"]))**order
        else:
            raise ConfigurationError(
                "{} calculation of equilibrium constant based on Gibbs energy "
                "is only supported for molarity or activity forms. "
                "Currently selected form: {}".format(rblock.name, c_form))

        # Check that heat of reaction is constant
        if config.heat_of_reaction is not constant_dh_rxn:
            raise ConfigurationError(
                "{} calculating equilibrium constants from Gibbs energy "
                "assumes constant heat of reaction. Please ensure you are "
                "using the constant_dh_rxn method for this reaction"
                .format(rblock.name))

        rblock.ds_rxn_ref = Var(
                doc="Specific molar entropy of reaction at reference state",
                units=units["entropy_mole"])
        set_param_from_config(rblock, param="ds_rxn_ref", config=config)

        rblock.T_eq_ref = Var(
                doc="Reference temperature for equilibrium constant",
                units=units["temperature"])
        set_param_from_config(rblock, param="T_eq_ref", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        units = rblock.parent_block().get_metadata().derived_units

        R = pyunits.convert(c.gas_constant, to_units=units["gas_constant"])

        return (exp(
            (-rblock.dh_rxn_ref / (R*T)) +
            (rblock.ds_rxn_ref / R)) * rblock._keq_units)

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        units = rblock.parent_block().get_metadata().derived_units
        R = pyunits.convert(c.gas_constant, to_units=units["gas_constant"])

        keq_val = value(exp(-rblock.dh_rxn_ref/(R*rblock.T_eq_ref) +
                            rblock.ds_rxn_ref/R) * rblock._keq_units)
        return 1/keq_val



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for Component objects

Author: Andrew Lee
"""
import pytest
import types

from pyomo.environ import ConcreteModel, Set, Param, Var, units as pyunits

from idaes.core.components import (Component, Solute, Solvent,
                                   Ion, Anion, Cation, Apparent)
from idaes.core.phases import (LiquidPhase, VaporPhase, SolidPhase, Phase,
                               PhaseType, AqueousPhase)
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
from idaes.core.property_meta import PropertyClassMetadata


class TestComponent():
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Component()
        m.comp2 = Component()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert m.comp.config.elemental_composition is None
        assert not m.comp.config._component_list_exists
        assert m.comp.config.henry_component is None
        assert m.comp.config.has_vapor_pressure

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp", "comp2"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        with pytest.raises(TypeError,
                           match="comp Generic Component objects do not "
                           "support is_solute\(\) method. Use a Solvent or "
                           "Solute Component instead."):
            m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        with pytest.raises(TypeError,
                           match="comp Generic Component objects do not "
                           "support is_solvent\(\) method. Use a Solvent or "
                           "Solute Component instead."):
            m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Component(default={
            "valid_phase_types": PhaseType.liquidPhase})

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Component(default={
            "valid_phase_types": PhaseType.vaporPhase})

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Component(default={
            "valid_phase_types": PhaseType.solidPhase})

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Component(default={
            "valid_phase_types": PhaseType.aqueousPhase})

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        # Generic components are never valid in the aqueous phase
        assert not m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Component(default={
            "valid_phase_types": [PhaseType.liquidPhase,
                                  PhaseType.vaporPhase]})

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_create_parameters(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)
        m.get_metadata().default_units["amount"] = pyunits.mol
        m.get_metadata().default_units["mass"] = pyunits.kg
        m.get_metadata().default_units["time"] = pyunits.s
        m.get_metadata().default_units["length"] = pyunits.m
        m.get_metadata().default_units["temperature"] = pyunits.K

        m.comp = Component(default={
            "parameter_data": {"mw": 10,
                               "pressure_crit": 1e5,
                               "temperature_crit": 500}})

        assert isinstance(m.comp.mw, Param)
        assert m.comp.mw.value == 10

        assert isinstance(m.comp.pressure_crit, Var)
        assert m.comp.pressure_crit.value == 1e5

        assert isinstance(m.comp.temperature_crit, Var)
        assert m.comp.temperature_crit.value == 500

    @pytest.mark.unit
    def test_create_parameters_convert(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)
        m.get_metadata().default_units["amount"] = pyunits.mol
        m.get_metadata().default_units["mass"] = pyunits.kg
        m.get_metadata().default_units["time"] = pyunits.s
        m.get_metadata().default_units["length"] = pyunits.m
        m.get_metadata().default_units["temperature"] = pyunits.K

        m.comp = Component(default={
            "parameter_data": {"mw": (10, pyunits.g/pyunits.mol),
                               "pressure_crit": (1, pyunits.bar),
                               "temperature_crit": (900, pyunits.degR)}})

        assert isinstance(m.comp.mw, Param)
        assert m.comp.mw.value == 1e-2

        assert isinstance(m.comp.pressure_crit, Var)
        assert m.comp.pressure_crit.value == 1e5

        assert isinstance(m.comp.temperature_crit, Var)
        assert m.comp.temperature_crit.value == 500


class TestSolute():
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Solute()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp"]

        assert isinstance(m.solute_set, Set)

        for j in m.solute_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Solute(default={
            "valid_phase_types": PhaseType.liquidPhase})

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Solute(default={
            "valid_phase_types": PhaseType.vaporPhase})

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Solute(default={
            "valid_phase_types": PhaseType.solidPhase})

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Solute(default={
            "valid_phase_types": PhaseType.aqueousPhase})

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        assert m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Solute(default={
            "valid_phase_types": [PhaseType.liquidPhase,
                                  PhaseType.vaporPhase]})

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)


class TestSovent():
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Solvent()

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.component_list, Set)

        for j in m.component_list:
            assert j in ["comp"]

        assert isinstance(m.solvent_set, Set)

        for j in m.solvent_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert not m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Solvent(default={
            "valid_phase_types": PhaseType.liquidPhase})

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Solvent(default={
            "valid_phase_types": PhaseType.vaporPhase})

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Solvent(default={
            "valid_phase_types": PhaseType.solidPhase})

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp6 = Solvent(default={
            "valid_phase_types": [PhaseType.liquidPhase,
                                  PhaseType.vaporPhase]})

        assert m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert m.comp6._is_phase_valid(m.Vap)
        assert not m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp7 = Solvent(default={
            "valid_phase_types": PhaseType.aqueousPhase})

        assert not m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert not m.comp7._is_phase_valid(m.Vap)
        assert m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)


class TestIon():
    @pytest.mark.unit
    def test_electrolye(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
                NotImplementedError,
                match="comp The IonData component class is intended as a base "
                "class for the AnionData and CationData classes, and should "
                "not be used directly"):
            m.comp = Ion(default={"_electrolyte": True})

    @pytest.mark.unit
    def test_not_electrolye(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
                PropertyPackageError,
                match="comp Ion Component types should only be used with "
                "Aqueous Phases"):
            m.comp = Ion(default={"_electrolyte": False})


class TestAnion():
    @pytest.mark.unit
    def test_not_electrolye(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
                PropertyPackageError,
                match="comp Ion Component types should only be used with "
                "Aqueous Phases"):
            m.comp = Anion(default={"_electrolyte": False})

    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.anion_set = Set()

        m.comp = Anion(default={"charge": -1, "_electrolyte": True})

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert "valid_phase_types" not in m.comp.config
        assert m.comp.config.charge == -1
        assert not m.comp.config._component_list_exists
        assert not m.comp.config.has_vapor_pressure
        with pytest.raises(ValueError):
            m.comp.config.has_vapor_pressure = True

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.anion_set, Set)

        for j in m.anion_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_invalid_charge(self, m):
        with pytest.raises(ConfigurationError,
                           match="an received invalid value for charge "
                           "configuration argument."
                           " Anions must have a negative charge."):
            m.an = Anion(default={"charge": +1, "_electrolyte": True})

    @pytest.mark.unit
    def test_no_charge(self, m):
        with pytest.raises(ConfigurationError,
                           match="an was not provided with a value "
                           "for charge."):
            m.an = Anion(default={"_electrolyte": True})

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(AttributeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert not m.comp._is_phase_valid(m.Liq)
        assert not m.comp._is_phase_valid(m.Sol)
        assert not m.comp._is_phase_valid(m.Vap)
        assert m.comp._is_phase_valid(m.Aqu)
        assert not m.comp._is_phase_valid(m.Phase)


class TestCation():
    @pytest.mark.unit
    def test_not_electrolye(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
                PropertyPackageError,
                match="comp Ion Component types should only be used with "
                "Aqueous Phases"):
            m.comp = Cation(default={"_electrolyte": False})

    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.cation_set = Set()

        m.comp = Cation(default={"charge": +1, "_electrolyte": True})

        return m

    @pytest.mark.unit
    def test_config(self, m):
        assert "valid_phase_types" not in m.comp.config
        assert m.comp.config.charge == +1
        assert not m.comp.config._component_list_exists
        assert not m.comp.config.has_vapor_pressure
        with pytest.raises(ValueError):
            m.comp.config.has_vapor_pressure = True

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m.cation_set, Set)

        for j in m.cation_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_invalid_charge(self, m):
        with pytest.raises(ConfigurationError,
                           match="cat received invalid value for charge "
                           "configuration argument."
                           " Cations must have a positive charge."):
            m.cat = Cation(default={"charge": -1, "_electrolyte": True})

    @pytest.mark.unit
    def test_no_charge(self, m):
        with pytest.raises(ConfigurationError,
                           match="cat was not provided with a value "
                           "for charge."):
            m.cat = Cation(default={"_electrolyte": True})

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(AttributeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert not m.comp._is_phase_valid(m.Liq)
        assert not m.comp._is_phase_valid(m.Sol)
        assert not m.comp._is_phase_valid(m.Vap)
        assert m.comp._is_phase_valid(m.Aqu)
        assert not m.comp._is_phase_valid(m.Phase)


class TestApparent():
    @pytest.fixture(scope="class")
    def m(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m._apparent_set = Set()

        m.comp = Apparent(default={"dissociation_species": {"comp": 1},
                                   "_electrolyte": True})

        return m

    @pytest.mark.unit
    def test_not_electrolye(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        m.comp = Apparent(default={"dissociation_species": {"comp": 1},
                                   "_electrolyte": False})

        for j in m.component_list:
            assert j in ["comp"]
        for j in m.solute_set:
            assert j in ["comp"]

        assert m.comp.config.dissociation_species == {"comp": 1}

    @pytest.mark.unit
    def test_no_dissociation_species(self):
        m = ConcreteModel()

        m.meta_object = PropertyClassMetadata()

        def get_metadata(self):
            return m.meta_object
        m.get_metadata = types.MethodType(get_metadata, m)

        with pytest.raises(
                ConfigurationError,
                match="dissoication_species argument was not set. "
                "Apparent components require the dissociation species to be "
                "defined."):
            m.comp = Apparent(default={"_electrolyte": False})

    @pytest.mark.unit
    def test_config(self, m):
        assert m.comp.config.valid_phase_types is None
        assert not m.comp.config._component_list_exists

    @pytest.mark.unit
    def test_populate_component_list(self, m):
        assert isinstance(m._apparent_set, Set)

        for j in m._apparent_set:
            assert j in ["comp"]

    @pytest.mark.unit
    def test_is_solute(self, m):
        assert m.comp.is_solute()

    @pytest.mark.unit
    def test_is_solvent(self, m):
        assert not m.comp.is_solvent()

    @pytest.mark.unit
    def test_is_phase_valid_no_assignment(self, m):
        with pytest.raises(TypeError):
            m.comp._is_phase_valid("foo")

    @pytest.mark.unit
    def test_is_phase_valid_liquid(self, m):
        m.comp3 = Apparent(default={
            "valid_phase_types": PhaseType.liquidPhase,
            "dissociation_species": {"comp": 1},
            "_electrolyte": True})

        m.Liq = LiquidPhase()
        m.Sol = SolidPhase()
        m.Vap = VaporPhase()
        m.Aqu = AqueousPhase()
        m.Phase = Phase()

        assert m.comp3._is_phase_valid(m.Liq)
        assert not m.comp3._is_phase_valid(m.Sol)
        assert not m.comp3._is_phase_valid(m.Vap)
        assert not m.comp3._is_phase_valid(m.Aqu)
        assert not m.comp3._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_vapor(self, m):
        m.comp4 = Apparent(default={
            "valid_phase_types": PhaseType.vaporPhase,
            "dissociation_species": {"comp": 1},
            "_electrolyte": True})

        assert not m.comp4._is_phase_valid(m.Liq)
        assert not m.comp4._is_phase_valid(m.Sol)
        assert m.comp4._is_phase_valid(m.Vap)
        assert not m.comp4._is_phase_valid(m.Aqu)
        assert not m.comp4._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_solid(self, m):
        m.comp5 = Apparent(default={
            "valid_phase_types": PhaseType.solidPhase,
            "dissociation_species": {"comp": 1},
            "_electrolyte": True})

        assert not m.comp5._is_phase_valid(m.Liq)
        assert m.comp5._is_phase_valid(m.Sol)
        assert not m.comp5._is_phase_valid(m.Vap)
        assert not m.comp5._is_phase_valid(m.Aqu)
        assert not m.comp5._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_aqueous(self, m):
        m.comp6 = Apparent(default={
            "valid_phase_types": PhaseType.aqueousPhase,
            "dissociation_species": {"comp": 1},
            "_electrolyte": True})

        assert not m.comp6._is_phase_valid(m.Liq)
        assert not m.comp6._is_phase_valid(m.Sol)
        assert not m.comp6._is_phase_valid(m.Vap)
        assert m.comp6._is_phase_valid(m.Aqu)
        assert not m.comp6._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_phase_valid_LV(self, m):
        m.comp7 = Apparent(default={
            "valid_phase_types": [PhaseType.liquidPhase,
                                  PhaseType.vaporPhase],
            "dissociation_species": {"comp": 1},
            "_electrolyte": True})

        assert m.comp7._is_phase_valid(m.Liq)
        assert not m.comp7._is_phase_valid(m.Sol)
        assert m.comp7._is_phase_valid(m.Vap)
        assert not m.comp7._is_phase_valid(m.Aqu)
        assert not m.comp7._is_phase_valid(m.Phase)

    @pytest.mark.unit
    def test_is_aqueous_phase_valid(self, m):
        assert m.comp._is_aqueous_phase_valid()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test for skeleton unit model.
"""

import pytest
from pyomo.environ import ConcreteModel, Constraint, Var, Set, value, \
    SolverStatus, TerminationCondition
from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models import SkeletonUnitModel
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import get_solver
from idaes.core.util.exceptions import ConfigurationError

__author__ = "Jaffer Ghouse"


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()
# -----------------------------------------------------------------------------


class TestSkeletonDefault(object):
    """
    Test for SkeletonUnitModel without callback for initialization

    """
    @pytest.fixture(scope="class")
    def skeleton_default(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.skeleton = SkeletonUnitModel(default={"dynamic": False})
        m.fs.skeleton.comp_list = Set(initialize=["c1", "c2"])

        # input vars for skeleton
        m.fs.skeleton.flow_comp_in = \
            Var(m.fs.time, m.fs.skeleton.comp_list, initialize=1.0)
        m.fs.skeleton.temperature_in = Var(m.fs.time, initialize=298.15)
        m.fs.skeleton.pressure_in = Var(m.fs.time, initialize=101)

        # output vars for skeleton
        m.fs.skeleton.flow_comp_out = \
            Var(m.fs.time, m.fs.skeleton.comp_list, initialize=1.0)
        m.fs.skeleton.temperature_out = Var(m.fs.time, initialize=298.15)
        m.fs.skeleton.pressure_out = Var(m.fs.time, initialize=101)

        # Surrogate model equations
        # Flow equation

        def rule_flow(m, t, i):
            return m.flow_comp_out[t, i] == m.flow_comp_in[t, i]

        m.fs.skeleton.eq_flow = \
            Constraint(m.fs.time, m.fs.skeleton.comp_list, rule=rule_flow)

        m.fs.skeleton.eq_temperature = Constraint(
            expr=m.fs.skeleton.temperature_out[0] ==
            m.fs.skeleton.temperature_in[0] + 10)
        m.fs.skeleton.eq_pressure = Constraint(
            expr=m.fs.skeleton.pressure_out[0] ==
            m.fs.skeleton.pressure_in[0] - 10)

        return m

    @pytest.mark.unit
    def test_ports(self, skeleton_default):

        inlet_dict = \
            {"flow_mol_comp": skeleton_default.fs.skeleton.flow_comp_in,
             "temperature": skeleton_default.fs.skeleton.temperature_in,
             "pressure": skeleton_default.fs.skeleton.pressure_in}
        outlet_dict = \
            {"flow_mol_comp": skeleton_default.fs.skeleton.flow_comp_out,
             "temperature": skeleton_default.fs.skeleton.temperature_out,
             "pressure": skeleton_default.fs.skeleton.pressure_out}

        skeleton_default.fs.skeleton.add_ports(
            name="inlet", member_dict=inlet_dict)
        skeleton_default.fs.skeleton.add_ports(
            name="outlet", member_dict=outlet_dict)

        # Check inlet port and port members
        assert hasattr(skeleton_default.fs.skeleton, "inlet")
        assert len(skeleton_default.fs.skeleton.inlet.vars) == 3
        assert hasattr(skeleton_default.fs.skeleton.inlet, "flow_mol_comp")
        assert hasattr(skeleton_default.fs.skeleton.inlet, "temperature")
        assert hasattr(skeleton_default.fs.skeleton.inlet, "pressure")

        # Check outlet port and port members
        assert hasattr(skeleton_default.fs.skeleton, "outlet")
        assert len(skeleton_default.fs.skeleton.outlet.vars) == 3
        assert hasattr(skeleton_default.fs.skeleton.outlet, "flow_mol_comp")
        assert hasattr(skeleton_default.fs.skeleton.outlet, "temperature")
        assert hasattr(skeleton_default.fs.skeleton.outlet, "pressure")

    @pytest.mark.unit
    def test_build(self, skeleton_default):
        # Check build of model
        assert hasattr(skeleton_default.fs.skeleton, "flow_comp_in")
        assert hasattr(skeleton_default.fs.skeleton, "temperature_in")
        assert hasattr(skeleton_default.fs.skeleton, "pressure_in")

        assert hasattr(skeleton_default.fs.skeleton, "eq_flow")

        assert hasattr(skeleton_default.fs.skeleton, "flow_comp_out")
        assert hasattr(skeleton_default.fs.skeleton, "temperature_out")
        assert hasattr(skeleton_default.fs.skeleton, "pressure_out")

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_exception(self, skeleton_default):
        # Check default initialize method raises ConfigurationError

        with pytest.raises(
            ConfigurationError,
            match="Degrees of freedom is not zero during start of "
                "initialization. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."):
            skeleton_default.fs.skeleton.initialize()

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_default_initialize(self, skeleton_default):

        skeleton_default.fs.skeleton.inlet.flow_mol_comp[0, "c1"].fix(2)
        skeleton_default.fs.skeleton.inlet.flow_mol_comp[0, "c2"].fix(2)
        skeleton_default.fs.skeleton.inlet.temperature.fix(325)
        skeleton_default.fs.skeleton.inlet.pressure.fix(200)

        assert degrees_of_freedom(skeleton_default) == 0

        skeleton_default.fs.skeleton.initialize()

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, skeleton_default):
        results = solver.solve(skeleton_default)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, skeleton_default):

        assert value(skeleton_default.fs.skeleton.
                     outlet.flow_mol_comp[0, "c1"]) == \
            pytest.approx(2, abs=1e-3)
        assert value(skeleton_default.fs.skeleton.
                     outlet.flow_mol_comp[0, "c2"]) == \
            pytest.approx(2, abs=1e-3)

        assert value(skeleton_default.fs.skeleton.
                     outlet.temperature[0]) == \
            pytest.approx(335, abs=1e-3)
        assert value(skeleton_default.fs.skeleton.
                     outlet.pressure[0]) == \
            pytest.approx(190, abs=1e-3)


class TestSkeletonCustom(object):
    """
    Test for SkeletonUnitModel with callback for initialization

    """
    @pytest.fixture(scope="class")
    def skeleton_custom(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.skeleton = SkeletonUnitModel(default={"dynamic": False})
        m.fs.skeleton.comp_list = Set(initialize=["c1", "c2"])

        # input vars for skeleton
        m.fs.skeleton.flow_comp_in = \
            Var(m.fs.time, m.fs.skeleton.comp_list, initialize=1.0)
        m.fs.skeleton.temperature_in = Var(m.fs.time, initialize=298.15)
        m.fs.skeleton.pressure_in = Var(m.fs.time, initialize=101)

        # output vars for skeleton
        m.fs.skeleton.flow_comp_out = \
            Var(m.fs.time, m.fs.skeleton.comp_list, initialize=1.0)
        m.fs.skeleton.temperature_out = Var(m.fs.time, initialize=298.15)
        m.fs.skeleton.pressure_out = Var(m.fs.time, initialize=101)

        # Model equations
        # Flow equation

        def rule_flow(m, t, i):
            return m.flow_comp_out[t, i] == m.flow_comp_in[t, i]

        m.fs.skeleton.eq_flow = \
            Constraint(m.fs.time, m.fs.skeleton.comp_list, rule=rule_flow)

        m.fs.skeleton.eq_temperature = Constraint(
            expr=m.fs.skeleton.temperature_out[0] ==
            m.fs.skeleton.temperature_in[0] + 10)
        m.fs.skeleton.eq_pressure = Constraint(
            expr=m.fs.skeleton.pressure_out[0] ==
            m.fs.skeleton.pressure_in[0] - 10)

        return m

    @pytest.mark.unit
    def test_ports(self, skeleton_custom):

        inlet_dict = \
            {"flow_mol_comp": skeleton_custom.fs.skeleton.flow_comp_in,
             "temperature": skeleton_custom.fs.skeleton.temperature_in,
             "pressure": skeleton_custom.fs.skeleton.pressure_in}
        outlet_dict = \
            {"flow_mol_comp": skeleton_custom.fs.skeleton.flow_comp_out,
             "temperature": skeleton_custom.fs.skeleton.temperature_out,
             "pressure": skeleton_custom.fs.skeleton.pressure_out}

        skeleton_custom.fs.skeleton.add_ports(
            name="inlet", member_dict=inlet_dict)
        skeleton_custom.fs.skeleton.add_ports(
            name="outlet", member_dict=outlet_dict)

        # Check inlet port and port members
        assert hasattr(skeleton_custom.fs.skeleton, "inlet")
        assert len(skeleton_custom.fs.skeleton.inlet.vars) == 3
        assert hasattr(skeleton_custom.fs.skeleton.inlet, "flow_mol_comp")
        assert hasattr(skeleton_custom.fs.skeleton.inlet, "temperature")
        assert hasattr(skeleton_custom.fs.skeleton.inlet, "pressure")

        # Check outlet port and port members
        assert hasattr(skeleton_custom.fs.skeleton, "outlet")
        assert len(skeleton_custom.fs.skeleton.outlet.vars) == 3
        assert hasattr(skeleton_custom.fs.skeleton.outlet, "flow_mol_comp")
        assert hasattr(skeleton_custom.fs.skeleton.outlet, "temperature")
        assert hasattr(skeleton_custom.fs.skeleton.outlet, "pressure")

    @pytest.mark.unit
    def test_build(self, skeleton_custom):
        # Check build of model
        assert hasattr(skeleton_custom.fs.skeleton, "flow_comp_in")
        assert hasattr(skeleton_custom.fs.skeleton, "temperature_in")
        assert hasattr(skeleton_custom.fs.skeleton, "pressure_in")

        assert hasattr(skeleton_custom.fs.skeleton, "eq_flow")

        assert hasattr(skeleton_custom.fs.skeleton, "flow_comp_out")
        assert hasattr(skeleton_custom.fs.skeleton, "temperature_out")
        assert hasattr(skeleton_custom.fs.skeleton, "pressure_out")

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_exception(self, skeleton_custom):
        # Check default initialize method raises ConfigurationError
        # when config argument for custom_initializer is True but a method
        # was not set to the attribute custom_initializer
        with pytest.raises(
            ConfigurationError,
            match="Degrees of freedom is not zero during start of "
                "initialization. Fix/unfix appropriate number of variables "
                "to result in zero degrees of freedom for initialization."):
            skeleton_custom.fs.skeleton.initialize()

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_custom_initialize(self, skeleton_custom):
        # Check default initialize method raises ConfigurationError
        # when config argument for custom_initializer is True but a method
        # was not set to the attribute custom_initializer
        with pytest.raises(ConfigurationError):
            skeleton_custom.fs.skeleton.initialize()

        skeleton_custom.fs.skeleton.inlet.flow_mol_comp[0, "c1"].fix(2)
        skeleton_custom.fs.skeleton.inlet.flow_mol_comp[0, "c2"].fix(2)
        skeleton_custom.fs.skeleton.inlet.temperature.fix(325)
        skeleton_custom.fs.skeleton.inlet.pressure.fix(200)

        assert degrees_of_freedom(skeleton_custom) == 0

        def my_initialize(unit, **kwargs):
            # Callback for user provided initialization sequence
            unit.eq_temperature.deactivate()
            unit.eq_pressure.deactivate()
            unit.outlet.temperature.fix(325)
            unit.outlet.pressure.fix(200)

            if degrees_of_freedom == 0:
                solver.solve(unit)

            unit.eq_temperature.activate()
            unit.eq_pressure.activate()
            unit.outlet.temperature.unfix()
            unit.outlet.pressure.unfix()

        skeleton_custom.fs.skeleton.config.initializer = my_initialize
        skeleton_custom.fs.skeleton.initialize()

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, skeleton_custom):
        results = solver.solve(skeleton_custom)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, skeleton_custom):

        assert value(skeleton_custom.fs.skeleton.
                     outlet.flow_mol_comp[0, "c1"]) == \
            pytest.approx(2, abs=1e-3)
        assert value(skeleton_custom.fs.skeleton.
                     outlet.flow_mol_comp[0, "c2"]) == \
            pytest.approx(2, abs=1e-3)

        assert value(skeleton_custom.fs.skeleton.
                     outlet.temperature[0]) == \
            pytest.approx(335, abs=1e-3)
        assert value(skeleton_custom.fs.skeleton.
                     outlet.pressure[0]) == \
            pytest.approx(190, abs=1e-3)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Reference state sub-methods for eNRTL activity coefficient method.

Only applicable to liquid/electrolyte phases

Reference state classes should implement the following methods:
    - ref_state : expressions for x and X at the reference state
    - ndIdn : derivative term for long-range contribution

Reference:

Song, Y. and Chen, C.-C., Symmetric Electrolyte Nonrandom Two-Liquid Activity
Coefficient Model, Ind. Eng. Chem. Res., 2009, Vol. 48, pgs. 7788–7797
"""
from pyomo.environ import Expression

from idaes.generic_models.properties.core.generic.utility import (
    get_component_object as cobj)
import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


# Near-zero value to use for unsymmetric reference state
EPS = 1e-20


class Unsymmetric(object):
    """
    Sub-methods for the symmetric (fused-salt) reference state
    """
    @staticmethod
    def ref_state(b, pname):
        def rule_x_ref(b, i):
            if i in b.params.solvent_set or i in b.params.solute_set:
                # Eqn 66
                return (b.mole_frac_phase_comp_true[pname, i] /
                        sum(b.mole_frac_phase_comp_true[pname, j]
                            for j in b.params.solvent_set|b.params.solute_set))
            else:
                return EPS

        b.add_component(pname+"_x_ref",
                        Expression(b.params.true_species_set, rule=rule_x_ref))

    @staticmethod
    def ndIdn(b, pname, i):
        # Eqn 71
        return 0


class Symmetric(object):
    """
    Sub-methods for the symmetric (fused-salt) reference state
    """
    @staticmethod
    def ref_state(b, pname):
        def rule_x_ref(b, i):
            if i in b.params.ion_set:
                # Eqn 66
                return (b.mole_frac_phase_comp_true[pname, i] /
                        sum(b.mole_frac_phase_comp_true[pname, j]
                            for j in b.params.ion_set))
            else:
                return 0

        b.add_component(pname+"_x_ref",
                        Expression(b.params.true_species_set, rule=rule_x_ref))

    @staticmethod
    def ndIdn(b, pname, i):
        # Eqn 75
        return 0.5*sum(cobj(b, j).config.charge**2*ndxdn(b, pname, i, j)
                       for j in b.params.ion_set)


def ndxdn(b, pname, i, j):
    x0 = getattr(b, pname+"_x_ref")

    # Delta function used in Eqns 73-76 (not defined in paper)
    if i == j:
        delta = 1
    else:
        delta = 0

    # Eqn 76
    return ((delta - x0[j]) /
            sum(b.mole_frac_phase_comp_true[pname, k]
                for k in b.params.ion_set))



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

import logging
import pytest
import sys

from io import StringIO
from pyomo.common.log import LoggingIntercept
from idaes.beta import import_beta

@pytest.mark.unit
def test_beta_module_exception():
    with pytest.raises(
            ImportError, match=r"Module 'idaes.tests.beta_mod' is in beta "
            "and must be imported using idaes.beta.import_beta\(\)."):
        import idaes.tests.beta_mod

@pytest.mark.unit
def test_beta_module_import():
    os = StringIO()
    with LoggingIntercept(os, 'idaes', logging.INFO):
        mod = import_beta('idaes.tests.beta_mod')
    assert mod.__name__ == 'idaes.tests.beta_mod'
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    os = StringIO()
    with LoggingIntercept(os, 'idaes', logging.INFO):
        mod = import_beta('.beta_mod')
    assert mod.__name__ == 'idaes.tests.beta_mod'
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    os = StringIO()
    with LoggingIntercept(os, 'idaes.logger', logging.INFO):
        mod = import_beta('idaes.logger')
    assert mod.__name__ == 'idaes.logger'
    assert os.getvalue().strip() == "Module 'idaes.tests.test_beta' "\
        "imported module 'idaes.logger' as a Beta module.  "\
        "This module is not declared beta and can be "\
        "imported using Python's normal import mechanisms."

@pytest.mark.unit
def test_beta_reimport():
    os = StringIO()
    with LoggingIntercept(os, 'idaes', logging.INFO):
        mod = import_beta('idaes.tests.beta_mod')
    assert mod.__name__ == 'idaes.tests.beta_mod'
    assert mod.reference_value == 42
    assert os.getvalue() == ""

    # Test that subsequent standard imports will still trigger an
    # exception, even if the module was previously imported using
    # import-beta()
    with pytest.raises(
            ImportError, match=r"Module 'idaes.tests.beta_mod' is in beta "
            "and must be imported using idaes.beta.import_beta\(\)."):
        import idaes.tests.beta_mod

    # But subsequent beta_import() modules return the same (original)
    # module object
    os = StringIO()
    with LoggingIntercept(os, 'idaes', logging.INFO):
        mod2 = import_beta('idaes.tests.beta_mod')
    assert mod2 is mod



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Test for Cappresse's module for NMPC.
"""

import pytest
from pyomo.environ import (Block, ConcreteModel,  Constraint, Expression,
                           Set, SolverFactory, Var, value, 
                           TransformationFactory, TerminationCondition)
from pyomo.network import Arc
from pyomo.common.collections import ComponentSet

from idaes.core import (FlowsheetBlock, MaterialBalanceType, EnergyBalanceType,
        MomentumBalanceType)
from idaes.core.util.tests.test_initialization import (
        AqueousEnzymeParameterBlock, EnzymeReactionParameterBlock,
        EnzymeReactionBlock)
from idaes.core.util.model_statistics import (degrees_of_freedom, 
        activated_equalities_generator)
from idaes.core.util.initialization import initialize_by_time_element
from idaes.core.util.exceptions import ConfigurationError
from idaes.generic_models.unit_models import CSTR, Mixer, MomentumMixingType
from idaes.apps.caprese import nmpc
from idaes.apps.caprese.nmpc import *
import idaes.logger as idaeslog

__author__ = "Robert Parker"


# See if ipopt is available and set up solver
if SolverFactory('ipopt').available():
    solver = SolverFactory('ipopt')
    solver.options = {'tol': 1e-6,
                      'mu_init': 1e-8,
                      'bound_push': 1e-8}
else:
    solver = None

def make_model(horizon=6, ntfe=60, ntcp=2, 
        inlet_E=11.91, inlet_S=12.92, steady=False, bounds=False):
    time_set = [0, horizon]

    m = ConcreteModel(name='CSTR model for testing')
    if steady:
        m.fs = FlowsheetBlock(default={'dynamic': False})
    else:
        m.fs = FlowsheetBlock(default={'dynamic': True,
                                       'time_set': time_set})

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(
            default={'property_package': m.fs.properties})
    m.fs.cstr = CSTR(default={'has_holdup': True,
                              'property_package': m.fs.properties,
                              'reaction_package': m.fs.reactions,
                              'material_balance_type': MaterialBalanceType.componentTotal,
                              'energy_balance_type': EnergyBalanceType.enthalpyTotal,
                              'momentum_balance_type': MomentumBalanceType.none,
                              'has_heat_of_reaction': True})

    m.fs.mixer = Mixer(default={
        'property_package': m.fs.properties,
        'material_balance_type': MaterialBalanceType.componentTotal,
        'momentum_mixing_type': MomentumMixingType.none,
        'num_inlets': 2,
        'inlet_list': ['S_inlet', 'E_inlet']})
    # Allegedly the proper energy balance is being used...

    # Time discretization
    if not steady:
        disc = TransformationFactory('dae.collocation')
        disc.apply_to(m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme='LAGRANGE-RADAU')

    # Fix geometry variables
    m.fs.cstr.volume[0].fix(1.0)

    # Fix initial conditions:
    if not steady:
        for p, j in m.fs.properties.phase_list*m.fs.properties.component_list:
            if j == 'Solvent':
                continue
            m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0.001)
    # Note: Model does not solve when initial conditions are empty tank

    m.fs.mixer.E_inlet.conc_mol.fix(0)
    m.fs.mixer.S_inlet.conc_mol.fix(0)

    for t, j in m.fs.time*m.fs.properties.component_list:
        if j == 'E':
            m.fs.mixer.E_inlet.conc_mol[t, j].fix(inlet_E)
        elif j == 'S':
            m.fs.mixer.S_inlet.conc_mol[t, j].fix(inlet_S)

    m.fs.mixer.E_inlet.flow_vol.fix(0.1)
    m.fs.mixer.S_inlet.flow_vol.fix(2.1)
    m.fs.mixer.E_inlet.conc_mol[:, 'Solvent'].fix(1.)
    m.fs.mixer.S_inlet.conc_mol[:, 'Solvent'].fix(1.)
    
    m.fs.mixer.E_inlet.temperature.fix(290)
    m.fs.mixer.S_inlet.temperature.fix(310)

    m.fs.inlet = Arc(source=m.fs.mixer.outlet, destination=m.fs.cstr.inlet)

    if not steady:
        # This constraint is in lieu of tracking the CSTR's level and allowing
        # the outlet flow rate to be another degree of freedom.
        # ^ Not sure how to do this in IDAES.
        @m.fs.cstr.Constraint(m.fs.time,
            doc='Total flow rate balance')
        def total_flow_balance(cstr, t):
            return (cstr.inlet.flow_vol[t] == cstr.outlet.flow_vol[t])

    # Specify initial condition for energy
    if not steady:
        m.fs.cstr.control_volume.energy_holdup[m.fs.time.first(), 'aq'].fix(300)

    TransformationFactory('network.expand_arcs').apply_to(m.fs)

    if bounds:
        m.fs.mixer.E_inlet.flow_vol.setlb(0.01)
        m.fs.mixer.E_inlet.flow_vol.setub(1.0)
        m.fs.mixer.S_inlet.flow_vol.setlb(0.5)
        m.fs.mixer.S_inlet.flow_vol.setub(5.0)

        m.fs.cstr.control_volume.material_holdup.setlb(0)
        holdup = m.fs.cstr.control_volume.material_holdup
        for t in m.fs.time:
            holdup[t, 'aq', 'S'].setub(20)
            holdup[t, 'aq', 'E'].setub(1)
            holdup[t, 'aq', 'P'].setub(5)
            holdup[t, 'aq', 'C'].setub(5)

    return m

if __name__ == '__main__':
    m_plant = make_model()
    m_controller = make_model(bounds=True)
    m_ss = make_model(steady=True)

    assert degrees_of_freedom(m_plant) == 0
    assert degrees_of_freedom(m_ss) == 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Test for element-by-element initialization on PID controller. This is important
as PID controllers have additional time-linking variables beyond derivative
and differential variables.
"""

import pytest
from pyomo.environ import (Block, ConcreteModel, Constraint, Expression,
                           Set, SolverFactory, Var, value, Param, Reals,
                           TransformationFactory, TerminationCondition,
                           exp)
from pyomo.network import Arc, Port
from pyomo.dae import DerivativeVar
from pyomo.common.collections import ComponentMap

from idaes.core import (FlowsheetBlock, 
                        MaterialBalanceType, 
                        EnergyBalanceType,
                        MomentumBalanceType, 
                        declare_process_block_class,
                        PhysicalParameterBlock,
                        StateBlock,
                        StateBlockData,
                        ReactionParameterBlock,
                        ReactionBlockBase,
                        ReactionBlockDataBase,
                        MaterialFlowBasis)
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.generic_models.unit_models import CSTR, Mixer, MomentumMixingType
from idaes.generic_models.control import PIDBlock, PIDForm
from idaes.core.util.initialization import initialize_by_time_element
from idaes.core.util.tests.test_initialization import \
        AqueousEnzymeParameterBlock, EnzymeReactionParameterBlock
from idaes.core.util import get_solver

import idaes.logger as idaeslog

__author__ = "Robert Parker"


solver = get_solver()


def make_model(horizon=6, ntfe=60, ntcp=2, inlet_E=11.91, inlet_S=12.92):

    time_set = [0, horizon]

    m = ConcreteModel(name='CSTR with level control')
    m.fs = FlowsheetBlock(default={'dynamic': True,
                                   'time_set': time_set})

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(
            default={'property_package': m.fs.properties})
    m.fs.cstr = CSTR(default={'has_holdup': True,
                              'property_package': m.fs.properties,
                              'reaction_package': m.fs.reactions,
                              'material_balance_type': MaterialBalanceType.componentTotal,
                              'energy_balance_type': EnergyBalanceType.enthalpyTotal,
                              'momentum_balance_type': MomentumBalanceType.none,
                              'has_heat_of_reaction': True})
    # MomentumBalanceType.none used because the property package doesn't
    # include pressure.

    m.fs.mixer = Mixer(default={
        'property_package': m.fs.properties,
        'material_balance_type': MaterialBalanceType.componentTotal,
        'momentum_mixing_type': MomentumMixingType.none,
        # MomentumMixingType.none used because the property package doesn't
        # include pressure.
        'num_inlets': 2,
       'inlet_list': ['S_inlet', 'E_inlet']})
    # Allegedly the proper energy balance is being used...

    # Time discretization
    disc = TransformationFactory('dae.collocation')
    disc.apply_to(m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme='LAGRANGE-RADAU')

    m.fs.pid = PIDBlock(default={'pv': m.fs.cstr.volume,
                                 'output': m.fs.cstr.outlet.flow_vol,
                                 'upper': 5.0,
                                 'lower': 0.5,
                                 'calculate_initial_integral': True,
                                 # ^ Why would initial integral be calculated
                                 # to be nonzero?
                                 'pid_form': PIDForm.velocity})

    m.fs.pid.gain.fix(-1.0)
    m.fs.pid.time_i.fix(0.1)
    m.fs.pid.time_d.fix(0.0)
    m.fs.pid.setpoint.fix(1.0)

    # Fix initial condition for volume:
    m.fs.cstr.volume.unfix()
    m.fs.cstr.volume[m.fs.time.first()].fix(1.0)

    # Fix initial conditions for other variables:
    for p, j in m.fs.properties.phase_list*m.fs.properties.component_list:
        if j == 'Solvent':
            continue
        m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0.001)
    # Note: Model does not solve when initial conditions are empty tank
    m.fs.cstr.control_volume.energy_holdup[m.fs.time.first(), 'aq'].fix(300)

    m.fs.mixer.E_inlet.conc_mol.fix(0)
    m.fs.mixer.S_inlet.conc_mol.fix(0)
    m.fs.mixer.E_inlet.conc_mol[:,'Solvent'].fix(1.)
    m.fs.mixer.S_inlet.conc_mol[:,'Solvent'].fix(1.)

    for t, j in m.fs.time*m.fs.properties.component_list:
        if j == 'E':
            m.fs.mixer.E_inlet.conc_mol[t, j].fix(inlet_E)
        elif j == 'S':
            m.fs.mixer.S_inlet.conc_mol[t, j].fix(inlet_S)

    m.fs.mixer.E_inlet.flow_vol.fix(0.1)
    m.fs.mixer.S_inlet.flow_vol.fix(2.1)

    # Specify a perturbation to substrate flow rate:
    for t in m.fs.time:
        if t < horizon/4:
            continue
        else:
            m.fs.mixer.S_inlet.flow_vol[t].fix(3.0)

    m.fs.mixer.E_inlet.temperature.fix(290)
    m.fs.mixer.S_inlet.temperature.fix(310)

    m.fs.inlet = Arc(source=m.fs.mixer.outlet, destination=m.fs.cstr.inlet)

    # Fix "initial condition" for outlet flow rate, as here it cannot be
    # specified by the PID controller
    m.fs.cstr.outlet.flow_vol[m.fs.time.first()].fix(2.2)

    TransformationFactory('network.expand_arcs').apply_to(m.fs)

    return m


@pytest.mark.component
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialize():
    '''Very rough test, just to make sure degrees of freedom are not violated.
    '''
    mod = make_model(horizon=2, ntfe=20, ntcp=1, inlet_E=11.91, inlet_S=12.92)

    assert degrees_of_freedom(mod) == 0

    originally_active = ComponentMap([(comp, comp.active) 
        for comp in mod.component_data_objects((Block, Constraint))])
    originally_fixed = ComponentMap([(var, var.fixed)
        for var in mod.component_data_objects(Var)])

    initialize_by_time_element(mod.fs, mod.fs.time, solver=solver, 
            outlvl=idaeslog.DEBUG,
            fix_diff_only=False)
    assert degrees_of_freedom(mod) == 0

    for comp in mod.component_data_objects((Block, Constraint)):
        assert comp.active == originally_active[comp]
    for var in mod.component_data_objects(Var):
        assert var.fixed == originally_fixed[var]


if __name__ == '__main__':
    test_initialize()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for DMF Jupyter "magics"
"""
# stdlib
from pathlib import Path
import os
from tempfile import TemporaryDirectory
from typing import Union
import webbrowser

# third-party
import pytest

# local
from idaes.dmf import magics, DMF
from idaes.dmf.magics import DMFMagicError
from idaes.dmf.resource import Resource

# monkey-patch webbrowser to do nothing
webbrowser.open_new = lambda url: None


class MockShell(object):
    """Mock object for IPython 'shell'.
    """

    def ev(self, name):
        """Mock evaluation.
        For most names, return the DMF class.
        For the special name "FAIL", raise an exception.
        For the special name "NODOCS", return a non-documented Python module.
        """
        if name == "FAIL":
            raise ValueError("Failure")
        elif name == "NODOCS":
            return os
        return DMF


@pytest.fixture
def magics_impl():
    shell = MockShell()
    return magics.DmfMagicsImpl(shell)


scratch_dir: Union[str, None] = None
scratch_path: Union[Path, None] = None


def setup_module(module):
    global scratch_dir, scratch_path
    scratch_dir = TemporaryDirectory(prefix="idaes_dmf_")  # easier to remove later
    scratch_path = Path(scratch_dir.name)


def teardown_module(module):
    global scratch_dir
    del scratch_dir


@pytest.mark.unit
def test_init_create(magics_impl):
    tmp_dir = scratch_path / "init_create"
    dmf = DMF(path=tmp_dir, create=True)
    magics_impl.dmf_init(str(tmp_dir), "create")
    assert magics_impl.initialized


@pytest.mark.unit
def test_init_existing(magics_impl):
    tmp_dir = scratch_path / "init_existing"
    dmf = DMF(path=tmp_dir, create=True)
    magics_impl.dmf_init(str(tmp_dir))


@pytest.mark.unit
def test_init_extraignored(magics_impl):
    tmp_dir = scratch_path / "init_extraignored"
    dmf = DMF(path=tmp_dir, create=True)
    magics_impl.dmf_init(str(tmp_dir), "foo")  # just generate warning
    magics_impl.dmf_init(str(tmp_dir), "foo", "bar")  # ditto


@pytest.mark.unit
def test_init_required(magics_impl):
    pytest.raises(DMFMagicError, magics_impl.dmf_status)
    pytest.raises(DMFMagicError, magics_impl.dmf_help, "anything")


@pytest.mark.unit
def test_init_badpath(magics_impl):
    nosuchpath = os.path.join(os.path.sep, *map(str, range(10)))
    pytest.raises(DMFMagicError, magics_impl.dmf_init, nosuchpath)
    pytest.raises(DMFMagicError, magics_impl.dmf_init, nosuchpath, "create")


@pytest.mark.unit
def test_init_goodpath(magics_impl):
    tmp_dir = scratch_path / "init_goodpath"
    magics_impl.dmf_init(str(tmp_dir), "create")


@pytest.mark.unit
def test_dmf_cmd(magics_impl):
    pytest.raises(DMFMagicError, magics_impl.dmf, "not a command")
    pytest.raises(DMFMagicError, magics_impl.dmf, "list stuff")
    magics_impl.dmf("")  # Empty value is OK


@pytest.mark.unit
def test_dmf_workspaces(magics_impl):
    tmp_dir = scratch_path / "dmf_workspaces"
    # Try with a good path, but no workspace
    # so should return "No valid workspaces found"
    msg = magics_impl.dmf_workspaces(tmp_dir)
    # make sure first token of response is not a number
    token = msg.split()[0]
    pytest.raises(ValueError, int, token)
    # Create a workspace
    magics_impl.dmf_init(tmp_dir, "create")
    # Try again, should return "1 workspace found" (or similar)
    msg = magics_impl.dmf_workspaces(tmp_dir)
    token = msg.split()[0]
    assert int(token) > 0


@pytest.mark.unit
def test_dmf_list(magics_impl):
    # Start before setting up DMF; should not work
    pytest.raises(DMFMagicError, magics_impl.dmf, "list")
    # Now set up the DMF, add 1 resource, call again. This time,
    # should be OK
    tmp_dir = scratch_path / "dmf_list"
    magics_impl.dmf_init(tmp_dir, "create")
    rsrc = Resource()
    magics_impl._dmf.add(rsrc)  # XXX: fragile since it accesses _dmf
    magics_impl.dmf("list")


@pytest.mark.unit
def test_dmf_status_initrequired(magics_impl):
    # should fail with no DMF
    pytest.raises(DMFMagicError, magics_impl.dmf_status)


@pytest.mark.unit
def test_dmf_status_topics(magics_impl):
    tmp_dir = scratch_path / "dmf_status_topics"
    # should fail with DMF, and topics (not implemented)
    magics_impl.dmf_init(str(tmp_dir), "create")
    pytest.raises(DMFMagicError, magics_impl.dmf_status, "uptime")
    # should succeed with DMF, and no topics
    magics_impl.dmf_status()


@pytest.mark.unit
def test_dmf_status_extra_meta(magics_impl):
    tmp_dir = scratch_path / "dmf_status_extra_meta"
    tmp_dir.mkdir()
    # by filling in the metadata, this will test the dmf_status code
    # that prints out list and dict data structures
    (tmp_dir / DMF.WORKSPACE_CONFIG).open("w").write(
        f"""
_id: this-is-a-temporary-config
logging:
    idaes.dmf.dmfbase:
        level: debug
        output: _stderr_
    root:
        output: _stdout_
    dmf:
        output: _stdout_
    .dmf.experiment:
        output: _stdout_
    # equivalent to previous
    idaes.dmf.experiment:
        output: {tmp_dir / 'experiment.log'}
    # user
    crazy.little.logger:
        level: error
        output: _stderr_
        """
    )
    magics_impl.dmf_init(str(tmp_dir))
    magics_impl.dmf_status()


@pytest.mark.unit
def test_dmf_help(magics_impl):
    tmp_dir = scratch_path / "dmf_help"
    magics_impl.dmf_init(str(tmp_dir), "create")
    magics_impl.dmf_help()
    # fail with bad args
    result = magics_impl.dmf_help("this", "that")
    assert result is None  # None means "OK"
    # with objects
    # a) "special" names
    for name in "dmf", "help", "idaes":
        magics_impl.dmf_help(name)
    # b) object
    magics_impl.dmf_help("idaes.dmf.dmfbase.DMF")
    # c) failure (still returns None)
    assert magics_impl.dmf_help("FAIL") is None




#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pyomo.environ as pyo
import idaes.core as idaes_core
from idaes.power_generation.unit_models.helm import HelmIsentropicTurbine
import idaes.core.util.convergence.convergence_base as cb
from idaes.generic_models.properties import iapws95
from idaes.core.util import get_solver
from pyomo.environ import units as pyunits
import idaes

def create_isentropic_turbine(f=1000, T_in=500, p_in=1e6, ratioP=0.7):
    m = pyo.ConcreteModel()
    m.fs = idaes_core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmIsentropicTurbine(default={"property_package": m.fs.properties})
    hin = iapws95.htpx(T_in*pyunits.K, p_in*pyunits.Pa) # J/mol
    m.fs.unit.inlet.flow_mol[0].fix(f)
    m.fs.unit.inlet.enth_mol[0].fix(hin)
    m.fs.unit.inlet.pressure[0].fix(p_in)
    m.fs.unit.ratioP[0].fix(ratioP)
    m.fs.unit.efficiency_isentropic.fix(0.9)
    m.fs.unit.initialize()
    return m

@cb.register_convergence_class("HelmIsentropicTurbine")
class HelmIsentropicTurbineConvergenceEvaluation(cb.ConvergenceEvaluation):
    def get_specification(self):
        """
        Returns the convergence evaluation specification for the
        PressureChanger unit model

        Returns
        -------
           ConvergenceEvaluationSpecification
        """
        s = cb.ConvergenceEvaluationSpecification()

        s.add_sampled_input(
                name='Inlet_Flowrate',
                pyomo_path='fs.unit.control_volume.properties_in[0].flow_mol',
                lower=1, upper=1e6, distribution="uniform")

        s.add_sampled_input(
                name='Inlet_Enthalpy',
                pyomo_path='fs.unit.control_volume.properties_in[0].enth_mol',
                lower=5000, upper=80000, distribution="uniform")

        s.add_sampled_input(
                name='Inlet_Pressure',
                pyomo_path='fs.unit.control_volume.properties_in[0].pressure',
                lower=1e5, upper=1e7, distribution="uniform")

        s.add_sampled_input(
                name='Pressure_Ratio',
                pyomo_path='fs.unit.ratioP[0]',
                lower=0.1, upper=1, distribution="uniform")

        s.add_sampled_input(
                name='Efficiency',
                pyomo_path='fs.unit.efficiency_isentropic[0]',
                lower=0.2, upper=1, distribution="uniform")

        return s

    def get_initialized_model(self):
        """
        Returns an initialized model for the PressureChanger unit model
        convergence evaluation

        Returns
        -------
           Pyomo model : returns a pyomo model of the PressureChanger unit
        """
        return create_isentropic_turbine()

    def get_solver(self):
        """
        Returns a solver object with the desired options for
        convergence evaluation (and initialization)

        Returns
        -------
           Pyomo solver
        """
        opt = get_solver(options={"max_iter": 25})
        return opt



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Test for plate heat exchanger model

Author: Paul Akula
"""
# Import Python libraries
import pytest

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, value, TerminationCondition

# Import IDAES Libraries
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
from idaes.power_generation.carbon_capture.mea_solvent_system.properties.liquid_prop \
    import LiquidParameterBlock
from idaes.power_generation.carbon_capture.mea_solvent_system.unit_models.phe \
    import PHE


solver = get_solver()


class TestPlateHeatExchanger:
    '''
    Tests for the plate heat exchange (PHE) model

    Inputs for states variables  are in SI units
    -Flowrate: mol/s
    -Temperature: K
    -Pressure: Pa
    '''

    @pytest.fixture(scope="module")
    def measurement(self):
        nccc_data = {'dataset1':
                     {'input':
                      {'hotside':
                       {'flowrate': 60.54879,
                        'temperature': 392.23,
                        'pressure': 202650,
                        'mole_fraction':
                        {'CO2': 0.0158,
                         'MEA': 0.1095,
                         'H2O': 0.8747}},
                       'coldside':
                       {'flowrate': 63.01910,
                        'temperature': 326.36,
                        'pressure': 202650,
                        'mole_fraction':
                        {'CO2': 0.0414,
                         'MEA': 0.1077,
                         'H2O': 0.8509}}},
                      'output':
                      {'hotside':
                       {'temperature': 330.42},
                       'coldside':
                       {'temperature': 384.91}}},
                     'dataset2':
                     {"input":
                      {'hotside':
                       {'flowrate': 102.07830,
                        'temperature': 389.57,
                        'pressure': 202650,
                        'mole_fraction':
                        {'CO2': 0.0284,
                         'MEA': 0.1148,
                         'H2O': 0.8568}},
                       'coldside':
                       {'flowrate': 104.99350,
                        'temperature': 332.26,
                        'pressure': 202650,
                        'mole_fraction':
                        {'CO2': 0.0438,
                         'MEA': 0.1137,
                         'H2O': 0.8425}}},
                      'output':
                      {'hotside':
                       {'temperature': 336.70},
                       'coldside':
                       {'temperature': 383.21}}}}

        return nccc_data

    @pytest.fixture(scope="class")
    def phe_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        # Set up property package
        m.fs.hotside_properties = LiquidParameterBlock()
        m.fs.coldside_properties = LiquidParameterBlock()

        # create instance of plate heat exchanger  on flowsheet
        m.fs.unit = PHE(default={'passes': 4,
                                 'channel_list': [12, 12, 12, 12],
                                 'divider_plate_number': 2,
                                 "hot_side": {
                                     "property_package": m.fs.hotside_properties
                                 },
                                 "cold_side": {
                                     "property_package": m.fs.coldside_properties
                                 }})
        return m

    @pytest.fixture(scope="class", params=['dataset1', 'dataset2'])
    def set_input_output(self, phe_model, measurement, request):
        for t in phe_model.fs.time:
            # hot fluid
            Th_out = measurement[request.param]['output']['hotside']['temperature']
            Th = measurement[request.param]['input']['hotside']['temperature']
            Fh = measurement[request.param]['input']['hotside']['flowrate']
            Ph = measurement[request.param]['input']['hotside']['pressure']
            xh_CO2 = measurement[request.param]['input']['hotside']['mole_fraction']['CO2']
            xh_MEA = measurement[request.param]['input']['hotside']['mole_fraction']['MEA']
            xh_H2O = measurement[request.param]['input']['hotside']['mole_fraction']['H2O']

            phe_model.fs.unit.hot_inlet.flow_mol[t].fix(Fh)
            phe_model.fs.unit.hot_inlet.temperature[t].fix(Th)
            phe_model.fs.unit.hot_inlet.pressure[t].fix(Ph)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "CO2"].fix(xh_CO2)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "H2O"].fix(xh_H2O)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "MEA"].fix(xh_MEA)

            # cold fluid
            Tc_out = measurement[request.param]['output']['coldside']['temperature']
            Tc = measurement[request.param]['input']['coldside']['temperature']
            Fc = measurement[request.param]['input']['coldside']['flowrate']
            Pc = measurement[request.param]['input']['coldside']['pressure']
            xc_CO2 = measurement[request.param]['input']['coldside']['mole_fraction']['CO2']
            xc_MEA = measurement[request.param]['input']['coldside']['mole_fraction']['MEA']
            xc_H2O = measurement[request.param]['input']['coldside']['mole_fraction']['H2O']
            phe_model.fs.unit.cold_inlet.flow_mol[t].fix(Fc)
            phe_model.fs.unit.cold_inlet.temperature[t].fix(Tc)
            phe_model.fs.unit.cold_inlet.pressure[t].fix(Pc)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "CO2"].fix(xc_CO2)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "H2O"].fix(xc_H2O)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "MEA"].fix(xc_MEA)

            output = {'hotside_temperature_expectation': Th_out,
                      'coldside_temperature_expectation': Tc_out}
            return output

    @pytest.mark.unit
    def test_phe_build(self, phe_model):
        assert phe_model.fs.unit.config.dynamic is False
        assert degrees_of_freedom(phe_model) == 12

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, phe_model):
        for t in phe_model.fs.time:
            # hot fluid
            phe_model.fs.unit.hot_inlet.flow_mol[t].fix(60.54879)
            phe_model.fs.unit.hot_inlet.temperature[t].fix(392.23)
            phe_model.fs.unit.hot_inlet.pressure[t].fix(202650)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "CO2"].fix(0.0158)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "H2O"].fix(0.8747)
            phe_model.fs.unit.hot_inlet.mole_frac_comp[t, "MEA"].fix(0.1095)

            # cold fluid
            phe_model.fs.unit.cold_inlet.flow_mol[t].fix(63.01910)
            phe_model.fs.unit.cold_inlet.temperature[t].fix(326.36)
            phe_model.fs.unit.cold_inlet.pressure[t].fix(202650)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "CO2"].fix(0.0414)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "H2O"].fix(0.1077)
            phe_model.fs.unit.cold_inlet.mole_frac_comp[t, "MEA"].fix(0.8509)
        initialization_tester(phe_model)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.integration
    def test_phe_validation(self, phe_model, set_input_output):

        phe_model.fs.unit.initialize()
        res = solver.solve(phe_model)

        assert res.solver.termination_condition == TerminationCondition.optimal

        # Mass conservation test
        assert abs(value(phe_model.fs.unit.hot_inlet.flow_mol[0] -
                         phe_model.fs.unit.hot_outlet.flow_mol[0])) <= 1e-6

        assert abs(value(phe_model.fs.unit.cold_inlet.flow_mol[0] -
                         phe_model.fs.unit.cold_outlet.flow_mol[0])) <= 1e-6

        # Energy conservation test
        assert value(phe_model.fs.unit.QH[0]) == value(phe_model.fs.unit.QC[0])

        # Exit tempreture prediction test
        assert value(phe_model.fs.unit.hot_outlet.temperature[0]) == \
            pytest.approx(
            set_input_output['hotside_temperature_expectation'], abs=1.5)

        assert value(phe_model.fs.unit.cold_outlet.temperature[0]) == \
            pytest.approx(
            set_input_output['coldside_temperature_expectation'], abs=1.5)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Compression Model.
Current model assumes a single vapor phase.
A future enhancement will add support for two-phase flow.

Reference:

Modekurti et al., (2017). "Design, Dynamic Modeling,
and Control of a Multistage CO2 Compresor System.
International Journal of Greenhouse Gas Control. v62. page 31-45

Created: April 2020

"""
__Author__ = "Quang Minh Le, John Eslick, Andrew Lee"

from enum import Enum
import pyomo.environ as pyo
from pyomo.environ import (
    NonNegativeReals,
    Var,
    value,
    log,
    sqrt,
    units as pyunits)
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class

from idaes.generic_models.unit_models.pressure_changer import CompressorData

import idaes.logger as idaeslog
from idaes.core.util import get_solver
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.constants import Constants as const

_log = idaeslog.getLogger(__name__)


class ImpellerType(Enum):
    cover_impeller = 1
    open_impeller = 2
    custom = 3


class VaneDiffuserType(Enum):
    vane_diffuser = 1
    vaneless_diffuser = 2
    custom = 3


def _build_cover_impeller_equations(blk):
    @blk.Constraint(blk.flowsheet().time)
    def impeller_work_coeff_eqn(b, t):
        phi = blk.mass_flow_coeff[t]
        return b.impeller_work_coeff[t] * 100 * phi == (
                          0.62 * phi - phi * (phi / 0.4)**3 + 0.0014) * 100

    @blk.Constraint(blk.flowsheet().time)
    def polytropic_head_coeff_vaned_diffuser_eqn(b, t):
        phi = blk.mass_flow_coeff[t]
        return b.mu_p_v[t] * 100 * phi == (
                      0.51 * phi + phi**2 - 7.6 * phi**3 - 0.00025) * 100


def _build_open_impeller_equations(blk):
    @blk.Constraint(blk.flowsheet().time)
    def impeller_work_coeff_eqn(b, t):
        phi = blk.mass_flow_coeff[t]
        return b.impeller_work_coeff[t] * phi == (
                      0.68 * phi - phi * (phi / 0.37)**3 + 0.002)

    @blk.Constraint(blk.flowsheet().time)
    def polytropic_head_coeff_vaned_diffuser_eqn(b, t):
        phi = blk.mass_flow_coeff[t]
        return b.mu_p_v[t] * phi == 0.5 * (1 + (phi - 0.065) / abs(
            phi - 0.065)) * (
                0.59 * phi + 0.7 * phi**2 - 7.5 * phi**3 - 0.00025
                ) + 0.5 * (1 - (phi - 0.065) / abs(phi - 0.065)) * 0.6 * phi


def _build_vane_diffuser_equations(blk):
    @blk.Constraint(blk.flowsheet().time)
    def polytropic_head_coeff_eqn(b, t):
        return b.mu_p[t] == b.mu_p_v[t]

    @blk.Constraint(blk.flowsheet().time)
    def polytropic_efficiency_eqn(b, t):
        return b.eff_p[t] == b.eff_p_v[t]


def _build_vaneless_diffuser_equations(blk):
    @blk.Constraint(blk.flowsheet().time)
    def polytropic_head_coeff_eqn(b, t):
        return b.mu_p[t] == b.impeller_work_coeff[t] * b.eff_p[t]

    @blk.Constraint(blk.flowsheet().time)
    def polytropic_efficiency_eqn(b, t):
        phi = blk.mass_flow_coeff[t]
        return (b.eff_p[t] - b.eff_p_v[t]) * (
                      0.04 + 5 * phi + b.eff_p_v[t]**3) == -0.017


@declare_process_block_class("CompressionStage", doc="Compression Stage Model")
class CompressionStageData(CompressorData):
    # Pressure changer with isentropic compressor options
    CONFIG = CompressorData.CONFIG()
    CONFIG.declare(
        "impeller_type",
        ConfigValue(
            default=None,
            domain=In(ImpellerType),
            description="Impeller type, if custom provide an expression rule",
            doc="""The type of impeller, if custom provide an expression rule
with the impeller_rule argument.
**default** - ImpellerType.open_impeller
**Valid values** - {
ImpellerType.cover_impeller,
ImpellerType.open_impeller,
ImpellerType.custom}""",
        ),
    )
    CONFIG.declare(
        "vane_diffuser_type",
        ConfigValue(
            default=None,
            domain=In(VaneDiffuserType),
            description="Vane diffuser type,"
            "if custom provide an expression rule",
            doc="""The type of vane diffuser,"
            "if custom provide an expression rule
with the vane_diffuser_rule argument.
**default** - VaneDiffuserType.vane_diffuser
**Valid values** - {
VaneDiffuserType.vane_diffuser,
VaneDiffuserType.vaneless_diffuser,
VaneDiffuserType.custom}""",
        ),
    )
    CONFIG.declare(
        "impeller_callback",
        ConfigValue(
            default=None,
            description="This is a callback that adds an impeller.  The"
            "callback function takes the impeller block data argument.",
        ),
    )
    CONFIG.declare(
        "vane_diffuser_callback",
        ConfigValue(
            default=None,
            description="This is a callback that adds a vane diffuser."
            "The callback function takes the vane diffuser"
            "block data argument.",
        ),
    )

    def build(self):
        super().build()

        if self.config.property_package.phase_list != ["Vap"]:
            raise ConfigurationError(
                "{} SCO2 compressor provided with invalid property package. "
                "The SCO2 compressor only support a single vapor phase."
                .format(self.name))

        # Some shorter refernces to property blocks
        properties_in = self.control_volume.properties_in
        properties_out = self.control_volume.properties_out

        self.mass_flow_coeff = Var(self.flowsheet().time,
                                   initialize=0.07,
                                   doc="Compressor Flow Coefficient",
                                   bounds=(0.01, 0.15))
        self.rspeed = Var(self.flowsheet().time,
                          initialize=70.0,
                          doc='Rotation Speed of The Impeller',
                          bounds=(0.5, 8000),
                          units=1 / pyunits.s)
        self.U2 = Var(self.flowsheet().time,
                      initialize=300,
                      doc="Impeller Tip Speed",
                      bounds=(0, 500),
                      units=pyunits.m / pyunits.s)
        self.delta_enth_polytropic = Var(self.flowsheet().time,
                                         initialize=2500,
                                         doc="Polytropic Enthalpy Change",
                                         units=pyunits.J / pyunits.mol,
                                         within=NonNegativeReals)
        self.impeller_work_coeff = Var(self.flowsheet().time,
                                       initialize=0.60,
                                       doc="Impeller Work Coefficient",
                                       bounds=(0.001, 1))
        self.mu_p_v = Var(self.flowsheet().time,
                          initialize=0.60,
                          doc="Polytropic Head Coefficient Vaned Diffuser",
                          bounds=(0.001, 1))
        self.mu_p = Var(self.flowsheet().time,
                        initialize=0.60,
                        doc="Polytropic Head Coefficient",
                        bounds=(0.001, 1))
        self.eff_p_v = Var(self.flowsheet().time,
                           initialize=0.85,
                           doc="Polytropic Efficiency Vaned Diffuser",
                           bounds=(0.001, 1))
        self.eff_p = Var(self.flowsheet().time,
                         initialize=0.85,
                         doc="Polytropic Efficiency",
                         bounds=(0.001, 1.0))
        self.Ma = Var(self.flowsheet().time,
                      initialize=0.10,
                      doc='Rotational Mach Number',
                      bounds=(0.5, 1.5))
        self.psi_3 = Var(self.flowsheet().time,
                         initialize=0.01,
                         doc="Dimensionless Exit Flow Coefficient",
                         within=NonNegativeReals,
                         units=pyunits.dimensionless)
        self.psi_s = Var(self.flowsheet().time,
                         initialize=0.01,
                         doc="Dimensionless Isentropic Head Coefficient",
                         within=NonNegativeReals,
                         units=pyunits.dimensionless)

        ########################################################
        # Declare variables for the model
        self.z_s = Var(initialize=0.97,
                       doc="Compressibility factors at suction")
        self.z_d1 = Var(initialize=0.88,
                        doc="Variable used for calculating"
                        "compressibility factor at discharge pressure")
        self.efficiency_mech = Var(initialize=0.97,
                                   doc="Mechanical Efficiency")
        self.eff_drive = Var(initialize=1.0,
                             doc="Driver efficiency")
        self.r2 = Var(initialize=0.05,
                      doc="Impeller Tip Radius",
                      units=pyunits.m)

        ############################################################

        self.ratioP[:] = 2.0   # make sure these have a number value
        self.deltaP[:] = 0     # to avoid an error later in initialize

        ###########################################################
        # Calculate Mach number and tip impeller speed
        @self.Constraint(self.flowsheet().time, doc="Mach Number")
        def Ma_con(b, t):
            # assume single vapor phase.
            speed_of_sound = self.control_volume.properties_in[
                t].speed_sound_phase['Vap']
            return b.Ma[t] == b.U2[t] / speed_of_sound

        @self.Constraint(self.flowsheet().time, doc="Rotation Speed "
                         "of the Impeller")
        def rspeed_con(b, t):
            return b.U2[t] == 2 * const.pi * b.r2 * b.rspeed[t]

        # set up the vane diffuser rule.
        vdcb = self.config.vane_diffuser_callback
        vdselect = self.config.vane_diffuser_type
        if vdselect is not VaneDiffuserType.custom and vdcb is not None:
            _log.warning("A valve diffuser type is not custom.")
        elif vdselect == VaneDiffuserType.vane_diffuser:
            _build_vane_diffuser_equations(self)
        elif vdselect == VaneDiffuserType.vaneless_diffuser:
            _build_vaneless_diffuser_equations(self)
        else:
            if vdcb is None:
                raise ConfigurationError(
                    "No custom vane diffuser callback provided")
            vdcb(self)

        # set up the impeller rule.
        icb = self.config.impeller_callback
        iselect = self.config.impeller_type
        if iselect is not ImpellerType.custom and icb is not None:
            _log.warning("An impeller type is not custom.")
        elif iselect == ImpellerType.cover_impeller:
            _build_cover_impeller_equations(self)
        elif iselect == ImpellerType.open_impeller:
            _build_open_impeller_equations(self)
        else:
            if icb is None:
                raise ConfigurationError(
                    "No custom impeller callback provided")
            icb(self)

        # efficiency for vane diffuser
        @self.Constraint(self.flowsheet().time, doc="Polytropic "
                         "Efficiency")
        def eff_p_v_cons(b, t):
            return b.eff_p_v[t] * b.impeller_work_coeff[t] == b.mu_p_v[t]

        # Total enthalpy and entropy change through compressor stage
        # Isentropic enthalpy change (hisen - hin)
        @self.Expression(self.flowsheet().time, doc="Specific "
                         "Enthalpy Change of Isentropic Process")
        def delta_enth_isentropic(b, t):
            return b.properties_isentropic[t].enth_mol - \
                properties_in[t].enth_mol

        # Total enthalpy change (ho - hi)
        @self.Expression(self.flowsheet().time, doc="Actual Enthalpy "
                         "Change")
        def delta_enth_actual(b, t):
            return properties_out[t].enth_mol - properties_in[t].enth_mol

        # Entropy change
        @self.Expression(self.flowsheet().time, doc="Entropy change in"
                         "Compressor Stage")
        def deltaS(b, t):
            return properties_out[t].entr_mol - properties_in[t].entr_mol

        @self.Constraint(self.flowsheet().time, doc="Polytropic "
                         "Enthalpy Correlation")
        def polytropic_correlation(b, t):
            return b.delta_enth_polytropic[t] == b.mu_p[t] * \
                b.delta_enth_actual[t] / b.impeller_work_coeff[t]

        # Correlation between actual enthalpy and polytropic enthalpy
        # This equation is obtained by Eq. (2.13) from Aungier textbook (2000)
        @self.Constraint(self.flowsheet().time, doc="Equation for"
                         "Polytropic Enthalpy")
        def delta_enth_polytropic_con(b, t):
            Tout = properties_out[t].temperature
            Tin = properties_in[t].temperature
            return b.delta_enth_polytropic[t] == b.delta_enth_actual[t] \
                - b.deltaS[t] * (Tout - Tin)/log(Tout / Tin)

        # Maximum impeller speed for individual stage
        @self.Expression(
            self.flowsheet().time,
            doc="Maximum Allowable Impeller Tip Speed")
        def U2max(b, t):
            phi = b.mass_flow_coeff[t]
            return sqrt((1984.1 * phi**2 - 616.88 * phi + 215.97) * 0.7 * 830)

        # Thermodynamic power
        @self.Expression(self.flowsheet().time,
                         doc="Thermodynamic Power")
        def fluid_pow(b, t):
            flow = properties_in[t].flow_mol
            return b.delta_enth_actual[t] * flow

        # Electricity power
        @self.Expression(self.flowsheet().time, doc="Shaft Power")
        def elec_pow(b, t):
            return b.fluid_pow[t] / (b.efficiency_mech * b.eff_drive)

        # -------------------------------------------------------------------
        # Static Volumetric Flow at Suction Inlet
        @self.Expression(self.flowsheet().time,
                         doc="Static Volumetric Flow at Suction Inlet")
        def Vsdot(b, t):
            flow_in = properties_in[t].flow_mol
            rho_v0 = properties_in[t].dens_mol
            return flow_in / rho_v0

        # Mass flow coefficient
        @self.Constraint(
            self.flowsheet().time, doc="Mass Flow Coefficient")
        def mass_flow_coeff_eqn(b, t):
            phi = b.mass_flow_coeff[t]
            return phi == b.Vsdot[t] / (const.pi * b.r2**2 * b.U2[t])

        # Static volumetric flow at impeller exit
        @self.Expression(self.flowsheet().time,
                         doc="Static volumetric flow at impeller exit")
        def V3dot(b, t):
            Tin = properties_in[t].temperature
            Tout = properties_out[t].temperature
            Pd = properties_out[t].pressure
            Ps = properties_in[t].pressure
            b.z_d = 0.5 * (b.z_s + b.z_d1)
            return b.Vsdot[t] * (b.z_d / b.z_s) * (Tout / Tin) * (Ps / Pd)

        # Dimensionless Exit Flow Coefficient (Performance Curve)
        @self.Constraint(self.flowsheet().time,
                         doc="Dimensionless Exit Flow Coefficient")
        def psi_3_eqn(b, t):
            b2 = 0.076 * 2 * b.r2
            return b.psi_3[t] == b.V3dot[t] / (
                const.pi * 2 * b.r2 * b2 * b.U2[t])

        # Isentropic head
        @self.Expression(self.flowsheet().time,
                         doc="Isentropic Head")
        def ys_model(b, t):
            k_v = self.control_volume.properties_in[t].heat_capacity_ratio
            a = (k_v - 1) / k_v
            Pd = properties_out[t].pressure
            Ps = properties_in[t].pressure
            Tin = properties_in[t].temperature
            mw = properties_in[t].mw
            gas_const = const.gas_constant
            return b.z_s * (gas_const / mw) * Tin * (1 / a) * ((Pd/Ps)**a - 1)

        # Dimensionless Isentropic Head Coefficient (Performance Curve)
        @self.Constraint(self.flowsheet().time,
                         doc="Dimensionless Isentropic Head Coefficient")
        def psi_s_eqn(b, t):
            return b.psi_s[t] == 2 * b.ys_model[t] / (b.U2[t]**2)

    def initialize(self, state_args=None, outlvl=idaeslog.NOTSET, solver=None,
                   optarg=None):
        """
        Initialize the inlet compressor stage model.
        This deactivates the specialized constraints,
        then does the isentropic compressor initialization,
        then reactivates the constraints and solves.

        Args:
            state_args (dict): Initial state for property initialization
            outlvl: Initialization logger level
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        unfix_ratioP = {}
        for t in self.flowsheet().time:
            # if there is not a good guess for efficiency or outlet pressure
            # provide something reasonable.
            eff = self.efficiency_isentropic[t]
            eff.fix(
                eff.value if pyo.value(eff) > 0.3 and pyo.value(
                    eff) < 1.0 else 0.85)
            # check for alternate pressure specs
            if self.outlet.pressure[t].fixed:
                self.ratioP[t] = pyo.value(
                    self.outlet.pressure[t] / self.inlet.pressure[t])
            elif self.control_volume.deltaP[t].fixed:
                self.ratioP[t] = pyo.value(
                    (self.control_volume.deltaP[t] + self.inlet.pressure[t]
                     ) / self.inlet.pressure[t])
            elif self.ratioP[t].fixed:
                self.outlet.pressure[t] = pyo.value(
                    self.ratioP[t] * self.inlet.pressure[t])
            else:
                if value(self.ratioP[t]) is None or value(
                        self.ratioP[t]) < 1.01:
                    self.ratioP[t].fix(1.5)
                else:
                    self.ratioP[t].fix()
                unfix_ratioP[t] = True

        # set list of constraints
        constraint_list = [
            "impeller_work_coeff_eqn",
            "polytropic_head_coeff_vaned_diffuser_eqn",
            "polytropic_head_coeff_eqn",
            "polytropic_efficiency_eqn",
            "Ma_con",
            "rspeed_con",
            "eff_p_v_cons",
            "polytropic_correlation",
            "delta_enth_polytropic_con",
            "mass_flow_coeff_eqn",
            "psi_3_eqn",
            "psi_s_eqn"]

        # deactivate unit model level constraints
        for c in constraint_list:
            self.component(c).deactivate()

        super().initialize(
            state_args=state_args, outlvl=outlvl, solver=solver, optarg=optarg)

        self.efficiency_isentropic.unfix()
        for t in self.flowsheet().time:
            if unfix_ratioP.get(t, False):
                self.ratioP[t].unfix()
        # Activate special constraints
        for c in constraint_list:
            getattr(self, c).activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info(
            "Initialization Complete: {}".format(idaeslog.condition(res))
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import sys
import os
sys.path.append(os.path.abspath('..')) # current folder is ~/examples
from idaes.apps.uncertainty_propagation.uncertainties import quantify_propagate_uncertainty
import pandas as pd
from idaes.apps.uncertainty_propagation.examples.rooney_biegler import rooney_biegler_model,rooney_biegler_model_opt

if __name__ == "__main__":
    variable_name = ['asymptote', 'rate_constant']
    data = pd.DataFrame(data=[[1,8.3],
                              [2,10.3],
                              [3,19.0],
                              [4,16.0],
                              [5,15.6],
                              [7,19.8]],
                        columns=['hour', 'y'])

    def SSE(model, data):
        expr = sum((data.y[i] - model.response_function[data.hour[i]])**2 for i in data.index)
        return expr


    results =  quantify_propagate_uncertainty(rooney_biegler_model,rooney_biegler_model_opt, data, variable_name, SSE)
